open Utils
open Ids
open Exprs
open Instruction
open Loop_nest
module CE = ConstExpr
module VE = VarExpr

module DI(Inst: Inst_t)(LN: Loopnest_t with module Inst := Inst)  = struct
  module RL = LN.Rev_loop_graph

  module DimMap = Make_map(struct type t = Dim.id let compare = Dim.compare_id end)

  type loop = U of int | V | S of int | A

  type bumped = (Index.t * Index.t) option

  type t = {dim: Dim.t;  incr: CE.t; last_tile_size: CE.t option;
            loop_scheme: loop list; is_vectorized: bool;is_closed:bool;
            index_clone: Index.gen; bump_info: bumped; next_index: Index.t;
            current_index: Index.t; indexes_list: (Index.t * Index.gen * (Index.t * VE.t) list) list }
  [@@deriving fields]

  (* A frozen snapshot of a dim state at some instant *)
  module Frozen = struct

    module T_arg = struct
      type t = {dim: Dim.t;  incr: CE.t; is_vectorized: bool;
                current_pack: Index.t;
                current_tile: Index.t; next_index: Index.t;
                current_index: Index.t; indexes_list: (Index.t * (Index.t * VE.t) list) list }
      [@@deriving eq]
    end

    module C_arg = struct
      type t = {dim: Dim.t; is_vectorized: bool; incr: CE.t;last_index: Index.t;
                last_tile: Index.t option; indexes_list: (Index.t * (Index.t * VE.t) list) list }
      [@@deriving eq]
    end

    type t = Nothing of Dim.t
           | VecUnr of Dim.t * CE.t * bool
           | Tile of T_arg.t
           | Close of C_arg.t

    let filter_indexes big_tile_index small_tile_index =
      let rec aux = function
        | [], l -> Some l
        | (ind1, _)::t1, (ind2, _)::t2 ->
          if Index.equal ind1 ind2 then aux (t1, t2)
          else None
        | _ -> None in
      (* indexes are sorted from bigger to smaller, reverse that so we can filter *)
      aux ((List.rev big_tile_index), (List.rev small_tile_index))
      |> Option.map List.rev

    (* This type allows us to have exhaustiveness check when we match on a pair that we assume to be well-formed *)
    type ordered_pair = Close_close of C_arg.t
                      | Close_tile of Index.t * C_arg.t * T_arg.t
                      | Close_vu of C_arg.t * CE.t * bool
                      | Close_nothing of C_arg.t
                      | Tile_tile of Index.t option * T_arg.t * T_arg.t
                      | Same_tile of T_arg.t
                      | Tile_vu of T_arg.t * CE.t * bool
                      | Tile_nothing of T_arg.t
                      | VU_VU of Dim.t * CE.t * CE.t * bool
                      | Same_VU of Dim.t * CE.t * bool
                      | VU_nothing of Dim.t * CE.t * bool
                      | Nothing_nothing of Dim.t

    (* returns Some ordered_pair if pairs are indeed well ordered, None elsewise *)
    let are_well_ordered t1 t2 = match t1, t2 with
      | Nothing d1, Nothing d2  ->
        Option.from_cond_val (Dim.equal d1 d2)
          (Nothing_nothing d1)
      | VecUnr (d1, incr, is_vectorized),
        Nothing d2 ->
        Option.from_cond_val (Dim.equal d1 d2)
          (VU_nothing (d1, incr, is_vectorized))
      | VecUnr (d1, incr1, is_v1),
        VecUnr (d2, incr2, is_v2) ->
        if (CE.equal incr1 incr2 && is_v1 = is_v2)
        then Some (Same_VU (d1, incr1, is_v1))
        else Option.from_cond_val (Dim.equal d1 d2 && (is_v1 = is_v2))
            (VU_VU (d1, incr1, incr2, is_v1))
      | Tile (T_arg.({dim =d2; _}) as t_arg),
        Nothing d1 ->
        Option.from_cond_val (Dim.equal d1 d2)
          (Tile_nothing t_arg)
      | Tile (T_arg.({dim =d2; _}) as t_arg),
        VecUnr (d1, incr, is_vec) ->
        Option.from_cond_val (Dim.equal d1 d2)
          (Tile_vu (t_arg, incr, is_vec))
      | T_arg.(Tile ({dim =d1; indexes_list = big_indxs; _} as t_big),
               Tile ({dim =d2;indexes_list = small_indxs;_} as t_small)) ->
        Option.bind (filter_indexes big_indxs small_indxs)
          (function
              ((ind, _)::_) ->
              Option.from_cond_val (Dim.equal d1 d2)
                (Tile_tile (Some ind, t_big, t_small))
            | [] ->
              if (T_arg.equal t_big t_small)
              then (Some (Same_tile t_big))
              else (Option.from_cond_val (Dim.equal d1 d2)
                      (Tile_tile (None, t_big, t_small)))
          )
      |  Close (C_arg.({dim =d2; _}) as c_arg),
         Nothing d1 ->
        Option.from_cond_val (Dim.equal d1 d2)
          (Close_nothing c_arg)
      | Close (C_arg.({dim =d2; _}) as c_arg),
        VecUnr (d1, incr, is_vec) ->
        Option.from_cond_val (Dim.equal d1 d2)
          (Close_vu (c_arg, incr, is_vec))
      | C_arg.(Close ({dim=d1; indexes_list= small_indxs;_} as c_big),
               Tile ({dim=d2; indexes_list= big_indxs;_} as t_small)) ->
        Option.bind (filter_indexes big_indxs small_indxs)
          (function
              ((ind, _)::_) ->
              Option.from_cond_val (Dim.equal d1 d2)
                (Close_tile (ind, c_big, t_small))
            | [] -> None
          )
      | Close  c1, Close c2 ->
        Option.from_cond_val (C_arg.equal c1 c2)
          (Close_close c1)
      | _ -> None

    exception Ill_ordered_tiles

    let gen_tile_loop = function
      | Nothing_nothing _ -> [], VE.zero, VE.zero, None
      | VU_nothing (_, _, _) ->
        [], VE.zero, VE.zero, None
      | Same_VU (dim, incr, is_vectorized) ->
        let increment = if  is_vectorized then (CE.const Inst.A.vec_size) else CE.one in
        let index = Index.from_dim dim in
        let tile_index = Index.map_prefix index (fun s -> s ^ "t") in
        let start = VE.zero in
        let halt = VE.(Infix.(min (start + const_expr incr) (size (Dim.size_id dim )))) in
        [tile_index], VE.index tile_index, VE.index tile_index,
        Some (RL.new_seq tile_index start halt increment)
      | VU_VU _ ->
        (* TODO handle this case *)
        assert false
      | Tile_nothing T_arg.({current_pack;_}) ->
        [], VE.index current_pack, VE.zero, None
      | Tile_vu (T_arg.({dim; current_pack; _}), incr, is_vectorized) ->
        let increment = if  is_vectorized then (CE.const Inst.A.vec_size) else CE.one in
        let index = Index.from_dim dim in
        let global_index = Index.map_prefix index (fun s -> s ^ "g") in
        let local_index = Index.map_prefix index (fun s -> s ^ "l") in
        let aux = [local_index, VE.zero] in
        let start = VE.index current_pack in
        let halt = VE.(Infix.(min (start + const_expr incr) (size (Dim.size_id dim )))) in
        [global_index; local_index], VE.index global_index, VE.index local_index,
        Some (RL.new_seq ~aux global_index start halt increment)
      | Close_nothing C_arg.({dim;_}) ->
        [], VE.index @@ Index.from_dim dim, VE.zero,
        None
      | Close_vu (C_arg.({dim; _}), incr, is_vectorized) ->
        let increment = if is_vectorized then (CE.const Inst.A.vec_size) else CE.one in
        let index = Index.from_dim dim in
        let all_index = Index.map_prefix index (fun s -> s ^ "all") in
        let local_index = Index.map_prefix index (fun s -> s ^ "l") in
        let aux = [local_index, VE.zero] in
        let start = VE.index index in
        let halt = VE.(Infix.(min (start + const_expr incr) (size (Dim.size_id dim )))) in
        [all_index; local_index], VE.index all_index, VE.index local_index,
        Some (RL.new_seq ~aux all_index start halt increment)
      | Tile_tile (Some tile_index, _, T_arg.({dim; is_vectorized; incr; current_index; _})) ->
        let increment = if is_vectorized then (CE.const Inst.A.vec_size) else CE.one in
        let local_index = Index.map_prefix current_index (fun s -> s ^ "l") in
        let glob_index = Index.map_prefix current_index (fun s -> s ^ "g")  in
        let aux = [local_index, VE.zero] in
        let start = VE.index tile_index in
        let halt = VE.(Infix.(min (start + const_expr incr) (size (Dim.size_id dim )))) in
        [glob_index; local_index], VE.index glob_index, VE.index local_index,
        Some (RL.new_seq ~aux glob_index start halt increment)
      | Tile_tile (None, _, T_arg.({dim; is_vectorized; incr; next_index; current_index; _})) ->
        let increment = if is_vectorized then (CE.const Inst.A.vec_size) else CE.one in
        let local_index = Index.map_prefix current_index (fun s -> s ^ "l") in
        let glob_index = Index.map_prefix current_index (fun s -> s ^ "g")  in
        let aux = [local_index, VE.zero] in
        let start = VE.index next_index in
        let halt = VE.(Infix.(min (start + const_expr incr) (size (Dim.size_id dim )))) in
        [glob_index; local_index], VE.index glob_index, VE.index local_index,
        Some (RL.new_seq ~aux glob_index start halt increment)
      | Same_tile T_arg.{dim; current_index; incr; is_vectorized;_} ->
        let increment = if is_vectorized then (CE.const Inst.A.vec_size) else CE.one in
        let tile_index = Index.map_prefix current_index (fun s -> s ^ "tile") in
        let start = VE.zero in
        let halt = VE.(Infix.(min (start + const_expr incr) (size (Dim.size_id dim )))) in
        [tile_index], VE.index tile_index, VE.index tile_index,
        Some (RL.new_seq tile_index start halt increment)
      | Close_tile (_, _, T_arg.({dim; is_vectorized;  incr;  current_index; _})) ->
        let increment = if is_vectorized then (CE.const Inst.A.vec_size) else CE.one in
        let local_index = Index.map_prefix current_index (fun s -> s ^ "l") in
        let glob_index = Index.map_prefix current_index (fun s -> s ^ "g")  in
        let aux = [local_index, VE.zero] in
        let start = VE.index current_index in
        let halt = VE.(Infix.(min (start + const_expr incr) (size (Dim.size_id dim )))) in
        [local_index; glob_index], VE.index glob_index, VE.index local_index,
        Some (RL.new_seq ~aux glob_index start halt increment)
      | Close_close C_arg.{last_index; is_vectorized; dim;_} ->
        let increment = if is_vectorized then (CE.const Inst.A.vec_size) else CE.one in
        let index = Index.map_prefix last_index (fun s -> s ^ "all") in
        let start = VE.zero in
        let halt = VE.(size @@ Dim.size_id dim) in
        [index], VE.index index, VE.index index, Some (RL.new_seq index start halt increment)

    let interval_loop_tile big_tile small_tile =
      let ordered_pair = Option.get_exn (are_well_ordered big_tile small_tile) Ill_ordered_tiles in
      gen_tile_loop ordered_pair
  end

  (* Get a frozen snapshot of the state of this dimension at a given moment *)
  let freeze dim = function
      Some {dim; incr; is_vectorized; is_closed; bump_info;
            current_index; next_index; indexes_list;_} ->
      let open Frozen in
      let indexes_list = List.map (fun (ind, _, expr_list) -> ind, expr_list) indexes_list in
      begin match bump_info with
        | Some (current_tile,  current_pack) ->
          if is_closed
          then
            Close {dim; incr; is_vectorized; last_tile = Some current_tile;
                   last_index = current_index; indexes_list}
          else
            Tile {dim; incr; is_vectorized; current_index; next_index;
                  current_pack; current_tile; indexes_list}
        | None ->
          if is_closed
          then
            Close {dim; incr; is_vectorized; last_tile = None;
                   last_index = current_index; indexes_list}
          else VecUnr (dim, incr, is_vectorized)
      end
    | None ->  Nothing dim

  let current_tile {bump_info;_} = Option.map fst bump_info
  let current_pack {bump_info;_} = Option.map snd bump_info

  let bump_index ({next_index; index_clone; bump_info; current_index; indexes_list; _} as elem) =
    let new_next = index_clone () in
    match bump_info with
    | None ->
      let current_pack, new_gen = Index.fresh_clone_mark current_index "p" () in
      let indexes_list = [current_index, new_gen, [current_pack, VE.zero]] in
      {elem with next_index = new_next; bump_info = Some (current_pack, current_pack);
                 indexes_list; current_index = next_index}
    | Some _ ->
      let fold_aux (new_cloned_index, acc) (ind, ind_clone, ind_list) =
        let new_clone = ind_clone () in
        let ind_list = (new_clone, VE.index new_cloned_index)::ind_list  in
        (new_clone, (ind, ind_clone, ind_list)::acc) in
      let base_new_index, new_gen = Index.fresh_clone_mark current_index "p" () in
      let last_cloned, folded_list = List.fold_left fold_aux (base_new_index, []) indexes_list in
      let indexes_list = (current_index, new_gen, [base_new_index, VE.zero])::(List.rev folded_list) in
      {elem with next_index = new_next; bump_info= Some (base_new_index, last_cloned);
                 indexes_list; current_index = next_index}

  let is_closed {is_closed;_} = is_closed

  let vectorize elem = assert (not (elem.is_vectorized));
    {elem with incr = CE.const Inst.A.vec_size; loop_scheme = V::elem.loop_scheme;
               indexes_list = []; is_vectorized = true}


  let close ({is_closed; dim; incr; indexes_list; current_index;_}as elem) = if is_closed
    then (Printf.printf "Dim %s is already closed\n" (Dim.show dim); assert false)
    else
      let clone = Index.clone current_index in
      let indexes_list = (current_index, clone, [])::indexes_list in
      let last_tile_size = Some incr in
      let incr = CE.size @@ Dim.size_id dim in
      {elem with indexes_list; last_tile_size; incr; is_closed = true}

  (* Apply a function to incr *)
  let map_incr f elem = {elem with incr = f elem.incr}

  let default dim =
    let current_index = Index.from_dim dim in
    let index_clone = Index.clone current_index in
    let next_index = index_clone () in
    {dim; incr = CE.zero; last_tile_size = None; current_index; next_index ; loop_scheme = []; index_clone;
     bump_info = None; is_closed = false; is_vectorized = false; indexes_list = [] ;
    }
end
