open Utils
open Ids
(* Module that factorizes all common code between ConstExpr and VarExpr - which
 * is not a lot for now, but we'll see *)
module Expr = struct
  (* Extensible type. This will be extend by both ConstExpr and VarExpr *)
  type t = ..

  type t +=
    | Const of int
    | One
    | Zero
    | Add of t * t
    | Sub of t * t
    | Mul of t * t
    | Min of t * t

   (* Each function of this module has a cont version. 
    * cont is a continuation that allows to give back control to modules that
    * have extended type t whenever we stumble upon a variant that is not
    * defined in this module.
    * This allows us to reuse the code of this module.
    * In other words, we are just building an open recursion mecanism.
    * *)

  let fail _ = failwith "Not a simple expr"

  let rec eq_cont cont e1 e2 = match e1, e2 with
    | Const c1, Const c2 -> Int.equal c1 c2
    | One, One  -> true
    | Zero, Zero  -> true
    | Add (e11, e12), Add (e21, e22) -> eq_cont cont e11 e21 && eq_cont cont e12 e22
    | Sub (e11, e12), Sub (e21, e22) -> eq_cont cont e11 e21 && eq_cont cont e12 e22
    | Mul (e11, e12), Mul (e21, e22) -> eq_cont cont e11 e21 && eq_cont cont e12 e22
    | Min (e11, e12), Min (e21, e22) -> eq_cont cont e11 e21 && eq_cont cont e12 e22
    | Const _, _ | _, Const _ | One, _ | _, One | Zero, _| _, Zero | 
      Add _, _ | _, Add _| Mul _, _  | _, Mul _ | Min _, _ | _, Min _  -> false
    | e1, e2 -> cont e1 e2


  let equal e1 e2 = eq_cont fail e1 e2

  (*  This recursor allows us to spare the burden of writing
   * | Add e1, e2 -> Add f e1, f e2
   * | Mul e1, e2 -> Mul f e1, f e2, etc. 
   * Therefore f always has type expr -> expr *)
  let rec recursor_cont cont f = function
    | (Const _) | One | Zero as expr -> f expr
    | Add (e1, e2) -> begin
        match  (recursor_cont cont f e1, recursor_cont cont f e2) with
          e1, Zero -> e1
        | Zero, e2 -> e2
        | e1, e2 -> Add (e1, e2)
      end
    | Sub (e1, e2) -> begin
        match  (recursor_cont cont f e1, recursor_cont cont f e2) with
          e1, Zero -> e1
        | e1, e2 -> Sub (e1, e2)
      end
    | Mul (e1, e2) ->
      begin
        match  (recursor_cont cont f e1, recursor_cont cont f e2) with
          _, Zero -> Zero
        | Zero, _ -> Zero
        | e1, One -> e1
        | One, e2 -> e2
        | e1, e2 -> Mul (e1, e2)
      end
    | Min (e1, e2) -> begin
        match (recursor_cont cont f e1, recursor_cont cont f e2) with
        | _, Zero | Zero, _ -> Zero
        | One, Const c | Const c, One -> if c < 1 then Zero else One
        | Const c1, Const c2 -> Const (min c1 c2)
        | e1, e2 -> Min (e1, e2)
      end

    | expr -> cont expr


  let recursor f = recursor_cont fail f


  (* Basically, return true if we can print it safely without any parentheses *)
  let is_atomic_cont cont = function
    | Const _ | One | Zero -> true
    | Add _ | Mul _ | Sub _ -> false
    | e -> cont e

  let is_atomic = is_atomic_cont fail

  (* Same trick than before with the continuation We need to write show
   * ourselves as ppx deriving is not supported for extensible type *)
  let rec show_cont shw_cont atomic_cont = function
    | Const i -> Format.sprintf "%d" i
    | Zero -> "0"
    | One -> "1"
    | Add (e1, e2) -> Format.sprintf "%s + %s" (show_cont shw_cont atomic_cont e1)
                         (show_cont shw_cont atomic_cont e2)
    | Sub (e1, e2) -> Format.sprintf "%s - %s" (show_cont shw_cont atomic_cont e1)
                         (show_cont shw_cont atomic_cont e2)
    | Mul (e1, e2) -> let show_paren e = if is_atomic_cont atomic_cont e
                        then show_cont shw_cont atomic_cont e
                        else Format.sprintf "(%s)" (show_cont shw_cont atomic_cont e) in
      Format.sprintf "%s * %s" (show_paren e1) (show_paren e2)
    | Min (e1, e2) ->  Format.sprintf "MIN(%s, %s)" (show_cont shw_cont atomic_cont e1)
                         (show_cont shw_cont atomic_cont e2)
    | e -> shw_cont e

  let show exp =
    show_cont fail fail exp

  let pp fmt expr = Format.fprintf fmt "%s" (show expr)

  let const i = match i with
    | 0 -> Zero
    | 1 -> One
    | _ -> Const i

  let one = One
  let zero = Zero

  let add e1 e2 = match e1, e2 with
    | Zero, e | e, Zero -> e
    | Const c1, Const c2 -> Const (c1 + c2)
    | _ -> Add (e1, e2)

  let sub e1 e2 = match e1, e2 with
    | e, Zero -> e
    | Const c1, Const c2 -> Const (c1 - c2)
    | _ -> Sub (e1, e2)


  let mul e1 e2 = match e1, e2 with
    | One, e | e, One -> e
    | Zero, _ | _, Zero -> Zero
    | Const c1, Const c2 -> Const (c1 * c2)
    | _ -> Mul (e1, e2)

  let min e1 e2 =
    match e1,  e2 with
    | _, Zero | Zero, _ -> Zero
    | One, Const c | Const c, One -> if c < 1 then Zero else One
    | Const c1, Const c2 -> Const (min c1 c2)
    | e1, e2 -> Min (e1, e2)

  let rec simplify_cont cont = function
    | Add (Zero, e) | Add (e, Zero) -> e
    | Mul (Zero, _) | Mul (_, Zero) -> Zero
    | Mul (One, e) | Mul (e, One) -> e
    | e -> recursor_cont cont (simplify_cont cont) e

  let simplify = simplify_cont fail
  module Infix = struct
    let (+) = add
    let ( * ) = mul
  end
end

(* Expression that is required to be statically evaluable
 * In practice, we add to Expr a SizeVar variant that represents
 * a parametric size in a program *)
module ConstExpr = struct
  include Expr

  type t +=
    | SizeVar of Size.id

  let rec recursor f = function
    | SizeVar _ as sv -> f sv
    | e -> Expr.recursor_cont (recursor f) f e

  let rec is_atomic = function
    | SizeVar _ -> true
    | e -> Expr.is_atomic_cont is_atomic e

  let size sid = SizeVar sid

  let rec equal ce1 ce2 = match ce1, ce2 with
    | SizeVar sid1, SizeVar sid2 -> Size.equal_id sid1 sid2
    | SizeVar _, _ | _, SizeVar _ -> false
    | _ -> Expr.eq_cont equal ce1 ce2

  let rec show cexpr = match cexpr with
    | Const _ | Zero | One | Add _ | Sub _ | Mul _ | Min _ -> Expr.show_cont show is_atomic cexpr
    | SizeVar ind_id -> Format.sprintf "%s" (Size.show_id ind_id)
    | _ -> assert false

  let pp fmt expr = Format.fprintf fmt "%s" (show expr)

  (* takes f that maps a size to another ConstExpr
   * (potentially living in another "space") *)
  let rec map_index f = function
    | SizeVar id -> f id
    | Const _ | One | Zero as c -> c
    | e -> recursor (map_index f) e

  let rec to_int_opt sizes_assoc cexpr =
    let open Opt_syntax in
    match cexpr with
    | SizeVar id -> List.assoc_opt id sizes_assoc
    | Const i -> Some i
    | One -> Some 1
    | Zero -> Some 0
    | Add (x, y) ->
      let+ x' = to_int_opt sizes_assoc x
      and+ y' = to_int_opt sizes_assoc y in
      x' + y'
    | Sub (x, y) ->
      let+ x' = to_int_opt sizes_assoc x
      and+ y' = to_int_opt sizes_assoc y in
      x' - y'
    | Mul (x, y) ->
      let+ x' = to_int_opt sizes_assoc x
      and+ y' = to_int_opt sizes_assoc y in
      x' * y'
    | Min (x, y) ->
      let+ x' = to_int_opt sizes_assoc x
      and+ y' = to_int_opt sizes_assoc y in
      Int.min x' y'
    | _ -> failwith "Not a Const Expr"
end

(* Expression with indexes *)
module VarExpr = struct
  include Expr

  type t +=
    | ConstExpr of ConstExpr.t
    | IndVar of Index.t

  let rec recursor f = function
    | ConstExpr _ | IndVar _ as e -> f e
    | e -> Expr.recursor_cont (recursor f) f e

  let rec equal ve1 ve2 = match ve1, ve2 with
    | IndVar vid1, IndVar vid2 -> Index.equal vid1 vid2
    | ConstExpr ce1, ConstExpr ce2 -> ConstExpr.equal ce1 ce2
    | IndVar _, _ |_, IndVar _| ConstExpr _, _ |  _, ConstExpr _ -> false
    | _ -> Expr.eq_cont equal ve1 ve2

  (* contain_var [var_expr] [var_id] returns true if Variable var_id is used in var_expr *)
  let rec contain_var var_expr var_id = match var_expr with
    | IndVar index -> Index.equal index var_id
    | Zero | One | Const _ | ConstExpr _ -> false
    | Add (e1, e2) | Sub (e1, e2) | Mul (e1, e2) | Min (e1, e2) ->
      contain_var e1 var_id || contain_var e2 var_id
    | _ -> assert false

  (* contain_var [var_expr] [var_id] returns true if Variable var_id is used in var_expr *)
  let rec access_dim_id var_expr dim = match var_expr with
    | IndVar index -> Dim.equal_id (Index.dim index) dim
    | Zero | One | Const _ | ConstExpr _ -> false
    | Add (e1, e2) | Sub (e1, e2) | Mul (e1, e2) | Min (e1, e2) ->
      access_dim_id e1 dim || access_dim_id e2 dim
    | _ -> assert false

  (* contain_var [var_expr] [var_id] returns true if Variable var_id is used in var_expr *)
  let access_dim var_expr dim =
    access_dim_id var_expr (Dim.id dim)

  let rec is_constant = function
    | IndVar _ -> false
    | Zero | One | Const _ | ConstExpr _ -> true
    | Add (e1, e2) | Sub (e1, e2) | Mul (e1, e2) | Min (e1, e2) ->  is_constant e1 && is_constant e2
    | _ -> assert false


  let rec is_atomic = function
    | ConstExpr cexpr -> ConstExpr.is_atomic cexpr
    | IndVar _ -> true
    | e -> Expr.is_atomic_cont is_atomic e

  let rec show expr = match expr with
    | Const _ | Zero | One | Add _ | Sub _ | Mul _ | Min _ -> Expr.show_cont show is_atomic expr
    | ConstExpr cexpr -> ConstExpr.show cexpr
    | IndVar index -> Format.sprintf "%s" (Index.show_id (Index.id index))
    | _ -> failwith "Not an IndExpr"

  let pp fmt expr = Format.fprintf fmt "%s" (show expr)

  let size size_id = ConstExpr (ConstExpr.size size_id )

  let index ind_id = IndVar ind_id

  let const_expr cexpr = match cexpr with
    | Const _ | Zero | One -> cexpr
    | _ -> ConstExpr cexpr

  let to_int_opt sizes_indexes = function
    | ConstExpr cexpr -> ConstExpr.to_int_opt sizes_indexes cexpr
    | _ -> None

  (* Apply a function on all indexes by recursing on an expression *)
  let rec map_cexpr f = function
    | ConstExpr ce  -> ConstExpr (f ce)
    | IndVar _ | Const _ | One | Zero as e -> e
    | e -> recursor (map_cexpr f) e

  (* Apply a function on all indexes by recursing on an expression *)
  let rec map_index f = function
    | IndVar vid -> f vid
    | ConstExpr _ | Const _ | One | Zero as e  ->  e
    | e -> recursor (map_index f) e

  let alpha_replace_pred pred var_expr expr =
    let replace = fun vid -> if pred vid then var_expr else IndVar vid in
    map_index replace expr

  let alpha_replace var_id var_expr expr =
      alpha_replace_pred (Index.equal var_id) var_expr expr

  let alpha_replace_dim dim var_expr expr =
      alpha_replace_pred (fun vid -> Dim.equal_id (Index.dim vid) (Dim.id dim)) var_expr expr

  let alpha_conv old_vid new_vid =
    alpha_replace old_vid (IndVar new_vid)

  let prop_const old_vid const expr =
    alpha_replace old_vid (Const const) expr
  module Infix = struct
    include Infix
    let ( +-- ) e ind = e + (index ind)
    let ( --+ ) ind e = index ind + e
    let ( *-- ) e ind = e * (index ind)
    let ( --* ) ind e = index ind * e
  end
end
