open Utils

module type GRAPH_TYPE = sig
  type leaf [@@deriving show]
  type node [@@deriving show]
end

(* We are going to assume that there is a Top - A node that is always superior to everything else *)
module G(G_type: GRAPH_TYPE) = struct
  include G_type
  type key = I of int | Top [@@deriving show, ord]

  let to_int = function
    | I i -> i
    | _ -> failwith "Not an internal"

  include G_type

  type elem = Base of leaf | Intern of node * key list [@@deriving show]
  type vertex = N of elem * key [@@deriving show]
  (* For now each node can only have one successor at most
   * Nodes also contains its children *)
  module MI = Map.Int
  type node_map = vertex MI.t

  let pp_node_map fmt map =
    map
    |> MI.bindings
    |> [%show: (int * vertex) list]
    |> Format.fprintf fmt "%s"
    
  (* We keep track of every max nodes - Nodes that have Top for successor *)
  type graph = {top_nodes: key list; graph:node_map} [@@deriving show]

  let empty = {top_nodes = []; graph = MI.empty}

  let find key {graph;_} = match key with
    | Top -> None
    | I i -> Some (MI.find i graph)

  let find_leaf key graph =
    match find key graph with
    | Some N (Base l, _) -> l
    | _ -> failwith "Not a leaf key"

  let add key node {top_nodes; graph} = match key with
    | Top -> failwith "Cannot add a node to Top"
    | I i -> {top_nodes; graph =MI.add i node graph}

  let modify key f {top_nodes; graph} = match key with
    | Top -> failwith "Cannot modify Top"
    | I i -> {top_nodes; graph =MI.modify i f graph}

  let top_elems {top_nodes;_} =
    top_nodes

  let ident = ref 0

  let key_gen () = I (Ref.post_incr ident)

  let elem = function
    | N (Base l, _) -> Either.Left l
    | N (Intern(n, children_list), _) -> Either.Right (n, children_list)

  let successor = function
    | N (_, key) -> key

  let set_successor succ_key = function
    N (node, Top) -> N(node, succ_key)
    | _ -> failwith "Tried to set successor of a node that is not max"

  let map_node f ({graph;_} as g) =
    let graph = MI.map (function N(Intern(n, l), next) -> N(Intern(f n, l), next)
                   | other -> other) graph in
    {g with graph}

  let map_base f ({graph;_} as g) =
    let graph = MI.map (function N(Base n, next) -> N(Base (f n), next)
                   | other -> other) graph in
    {g with graph}

  let modify_base key f ({graph;_} as g) =
    let graph = MI.modify (to_int key) (function N(Base n, next) -> N(Base (f n), next)
                   | _ -> failwith "Supposed to be a base node") graph in
    {g with graph}

  (* Not my proudest function, but this will do the job *)
  let modify_node key f ({graph;top_nodes} as g) =
    let vertex_f = function N(Intern (n, k), next) -> N (Intern( f n, k), next)
                          | _ -> failwith "Not an internal node" in
    match key with
    | Top -> {g with graph = List.fold_left (fun g k -> MI.modify (to_int k) vertex_f g) graph top_nodes}
    | I i ->
    let graph = MI.modify i vertex_f graph in
    {g with graph}


  (* Add a node on top of current graph *)
  let enclose node {top_nodes; graph}  =
    let new_key = key_gen () in
    let graph = List.fold_left
      (fun graph k ->
         MI.modify (to_int k) (set_successor new_key) graph)
      graph top_nodes
    |> MI.add (to_int new_key) (N (Intern (node, top_nodes), Top)) in
    new_key, {top_nodes = [new_key]; graph}

  let add_child at_begin parent_key child_key {top_nodes; graph} =
    let insert_child_key l =
      if at_begin
      then  child_key::l
      else l @ [child_key] in
    match parent_key with
    | Top ->
      { top_nodes = insert_child_key top_nodes; graph}
    | I parent_key ->
      let graph = MI.modify parent_key
          (function
              N (Intern(n, children_list), succ) ->
              N (Intern(n, insert_child_key children_list), succ)
            | _ -> failwith "Base cannot be a parent")
          graph  in
      {top_nodes; graph}

  let aux_insert_base base parent_key graph =
    let base = N (Base base, parent_key) in
    let new_key = key_gen () in
    let graph = add_child true parent_key new_key graph in
    new_key, add new_key base graph

  (* Insert a base node without parent *)
  let insert_base base graph =
    aux_insert_base base Top graph

  (* Hook a base node under a parent *)
  let hook_base base parent_key graph =
    aux_insert_base base parent_key graph

  (* Assumes there are no two nodes with the same key *)
  let embed at_begin {top_nodes =top_small; graph = small_graph} { top_nodes = top_big; graph = big_graph} hook_key =
    let merge_list l1 l2 =
      if at_begin then l1 @ l2
      else l2 @ l1 in
    match hook_key with
    | Top ->
      let graph = MI.merge (fun _ small_opt big_opt -> match small_opt, big_opt with
          | Some N (n, succ), None ->
            Some (N (n, succ))
          | None, Some N (n, succ) ->
            Some (N (n, succ))
          | _ -> failwith "Some duplication happened") small_graph big_graph in
      let top_nodes = merge_list top_small top_big in
      {graph; top_nodes }
    | I internal ->
      let graph = MI.merge (fun k l1_opt l2_opt -> match l1_opt, l2_opt with
          | Some N (n, Top), None ->
            Some (N (n, hook_key))
          | None, Some N (Intern(node, children), succ) when k = internal ->
            Some (N (Intern(node, merge_list top_small children), succ))
          | None, Some n  -> Some n
          | Some n, None  -> Some n
          | _ -> failwith "Some keys are duplicated") small_graph big_graph in
      {top_nodes = top_big; graph}

  let embed_before = embed true

  let embed_after = embed false

  type ('node, 'leaf) tree =
    | Top_loop of ('node, 'leaf) tree list
    | Node of 'node * ('node, 'leaf) tree list
    | Leaf of 'leaf

  let base_points graph =
    MI.filter_map  (fun _ -> function
          N (Base n, _) -> Some n
        | _ -> None) graph
    |> MI.keys |> List.of_enum


  let to_tree ({top_nodes;_} as graph) =
    let rec build_node key =
      let node = find key graph in
      match elem @@ Option.get node with
      | Either.Left leaf ->
        Leaf leaf
      | Either.Right (node, children) ->
        let children_list = List.map build_node children in
        Node (node, children_list) in
    let build_tree key_list = match key_list with
      | [key] ->
        build_node key
      | _ ->
        let node_list = List.map build_node key_list in
        Top_loop node_list in
    build_tree top_nodes
end
