open Ids
open Exprs
module VE = VarExpr
open Utils
module T = Tensor
open Arch

module type Inst_t = sig
  module A: Vec_arch_t

  type inst_type = Vec | Scal [@@deriving eq, show]

  type t =
    | Nop
    (* Allocated is a placeholder for "factorized out": instructions that are
     * in some way "out of scope", we don't expect to generate a fresh name for
     * them. Still uneasy with that though *)
    | Allocated of inst_type * string
    | Assign of t * inst_type * string
    | Read of T.t * T.accesses
    | Add of t * t
    | Sub of t * t
    | Mul of t * t
    | Write of t * T.t * T.accesses
    | Vread of T.t * T.accesses
    | Vadd of t * t
    | Vsub of t * t
    | Vmul of t * t
    | Vbcst of t
    | Vwrite of t * T.t * T.accesses [@@deriving eq, show]

  val vectorize_on_dim: Dim.id -> t -> t
  val map_expr: Dim.id -> (VE.t -> VE.t ) -> t -> t
  val access: Dim.id -> t -> bool
  val is_tensor_readonly: t list -> T.t -> bool

  type mem_accesses_inf = (T.t * T.accesses) * (inst_type * string)
  val fact_loads_on_dims:  Dim.id list -> mem_accesses_inf list  -> t list
    -> mem_accesses_inf list * t list * t list
  val fact_stores_on_dims:  Dim.id list -> mem_accesses_inf list  -> t list
    -> mem_accesses_inf list * t list * t list
  val swap_tensor: T.t -> T.t -> t list -> t list
  val swap_accesses: (T.t -> T.accesses -> T.accesses) -> t list -> t list
  val gen_code: string -> t list -> string list * string list * string
end

module I(V: Vec_arch_t): Inst_t = struct
  module A = V

  type inst_type = Vec | Scal [@@deriving eq, show]

  let string_of_inst_type = function
    | Vec -> "vec"
    | Scal -> "scal"

  type t =
    | Nop
    (* Allocated is a placeholder for "factorized out": instructions that are
     * in some way "out of scope", we don't expect to generate a fresh name for
     * them. Still uneasy with that though *)
    | Allocated of inst_type * string
    | Assign of t * inst_type * string
    | Read of T.t * T.accesses
    | Add of t * t
    | Sub of t * t
    | Mul of t * t
    | Write of t * T.t * T.accesses
    | Vread of T.t * T.accesses
    | Vadd of t * t
    | Vsub of t * t
    | Vmul of t * t
    | Vbcst of t
    | Vwrite of t * T.t * T.accesses [@@deriving eq, show]

  type 'a recurse =
    | NopF
    (* Allocated is a placeholder for "factorized out": instructions that are
     * in some way "out of scope", we don't expect to generate a fresh name for
     * them. Still uneasy with that though *)
    | AllocatedF of inst_type * string
    | AssignF of 'a * inst_type * string
    | ReadF of T.t * T.accesses
    | AddF of 'a * 'a
    | SubF of 'a * 'a
    | MulF of 'a * 'a
    | WriteF of 'a * T.t * T.accesses
    | VreadF of T.t * T.accesses
    | VaddF of 'a * 'a
    | VsubF of 'a * 'a
    | VmulF of 'a * 'a
    | VbcstF of 'a
    | VwriteF of 'a * T.t * T.accesses

  let map f inst = match inst with
    | AssignF (v, t, name) -> AssignF (f v, t, name)
    | AddF (v1, v2) -> AddF (f v1, f v2)
    | SubF (v1, v2) -> SubF (f v1, f v2)
    | MulF (v1, v2) -> MulF (f v1, f v2)
    | WriteF (v, tens, acc) -> WriteF (f v, tens, acc)
    | VaddF (v1, v2) -> VaddF (f v1, f v2)
    | VsubF (v1, v2) -> VsubF (f v1, f v2)
    | VmulF (v1, v2) -> VmulF (f v1, f v2)
    | VbcstF v -> VbcstF (f v)
    | VwriteF (v, tens, acc) -> VwriteF (f v, tens, acc)
    (* Base cases *)
    | NopF -> NopF
    | AllocatedF (t, name) -> AllocatedF (t, name)
    | ReadF (tens, acc) -> ReadF (tens, acc)
    | VreadF (tens, acc) -> VreadF (tens, acc)

  type ('a, 'b, 'c) classifier = No_op of 'c | Unop of 'a * ('b -> 'c) | Binop of 'a * 'a * ('b -> 'b -> 'c)

  let explode = function
    | AssignF (v, t, name) -> Unop (v, fun v -> Assign (v, t, name))
    | AddF (v1, v2) -> Binop (v1, v2, fun v1 v2 -> Add ( v1, v2))
    | SubF (v1, v2) -> Binop (v1, v2, fun v1 v2 -> Sub ( v1, v2))
    | MulF (v1, v2) -> Binop (v1, v2, fun v1 v2 -> Mul ( v1, v2))
    | WriteF (v, tens, acc) -> Unop (v, fun v -> Write (v, tens, acc))
    | VaddF (v1, v2) -> Binop (v1, v2, fun v1 v2 -> Vadd ( v1, v2))
    | VsubF (v1, v2) -> Binop (v1, v2, fun v1 v2 -> Vsub ( v1, v2))
    | VmulF (v1, v2) -> Binop (v1, v2, fun v1 v2 -> Vmul ( v1, v2))
    | VbcstF v -> Unop (v, fun v -> Vbcst v)
    | VwriteF (v, tens, acc) -> Unop (v, fun v -> Vwrite (v, tens, acc))
    (* Base cases *)
    | NopF -> No_op Nop
    | AllocatedF (t, name) -> No_op (Allocated (t, name))
    | ReadF (tens, acc) -> No_op (Read (tens, acc))
    | VreadF (tens, acc) -> No_op (Vread (tens, acc))

  module InstAlgebra = Algebra.A(struct type 'a t = 'a recurse let map = map end)

  let wrap = function
    | Assign (v, t, name) -> AssignF ( v, t, name)
    | Add (v1, v2) -> AddF ( v1,  v2)
    | Sub (v1, v2) -> SubF ( v1,  v2)
    | Mul (v1, v2) -> MulF ( v1,  v2)
    | Write (v, tens, acc) -> WriteF ( v, tens, acc)
    | Vadd (v1, v2) -> VaddF ( v1,  v2)
    | Vsub (v1, v2) -> VsubF ( v1,  v2)
    | Vmul (v1, v2) -> VmulF ( v1,  v2)
    | Vbcst v -> VbcstF ( v)
    | Vwrite (v, tens, acc) -> VwriteF (v, tens, acc)
    (* Base cases *)
    | Nop -> NopF
    | Allocated (t, name) -> AllocatedF (t, name)
    | Read (tens, acc) -> ReadF (tens, acc)
    | Vread (tens, acc) -> VreadF (tens, acc)

  let unwrap = function
    | AssignF (v, t, name) -> Assign ( v, t, name)
    | AddF (v1, v2) -> Add ( v1,  v2)
    | SubF (v1, v2) -> Sub ( v1,  v2)
    | MulF (v1, v2) -> Mul ( v1,  v2)
    | WriteF (v, tens, acc) -> Write ( v, tens, acc)
    | VaddF (v1, v2) -> Vadd ( v1,  v2)
    | VsubF (v1, v2) -> Vsub ( v1,  v2)
    | VmulF (v1, v2) -> Vmul ( v1,  v2)
    | VbcstF v -> Vbcst ( v)
    | VwriteF (v, tens, acc) -> Vwrite (v, tens, acc)
    (* Base cases *)
    | NopF -> Nop
    | AllocatedF (t, name) -> Allocated (t, name)
    | ReadF (tens, acc) -> Read (tens, acc)
    | VreadF (tens, acc) -> Vread (tens, acc)

  (* Allocated and Assign are supposed not to access any dimension, which is ... doubtful *)
  let access dim =
    InstAlgebra.hylo wrap
      (function
        | ReadF (tensor, accesses) | VreadF (tensor, accesses) -> T.does_access tensor accesses dim
        | WriteF (expr, tensor, accesses) | VwriteF (expr, tensor, accesses) ->
          expr || T.does_access tensor accesses dim
        | AddF (b1, b2)
        | MulF (b1, b2)
        | SubF (b1, b2)
        | VaddF (b1, b2)
        | VmulF (b1, b2)
        | VsubF (b1, b2) -> b1 || b2
        | AllocatedF _
        | NopF -> false
        | AssignF (b, _, _)
        | VbcstF b -> b)

  let map_expr dim f =
    let open InstAlgebra in
    let fn = function
    | ReadF(s, access_list) ->
      ReadF(s, List.modify_assoc Dim.equal_id (Option.map f) dim access_list)
    | WriteF(value, tensor, access_list) ->
      WriteF(value, tensor, List.modify_assoc Dim.equal_id (Option.map f) dim access_list)
    | VreadF(s, access_list) ->
      VreadF(s, List.modify_assoc Dim.equal_id (Option.map f) dim access_list)
    | VwriteF(value, tensor, access_list) ->
      VwriteF(value, tensor, List.modify_assoc Dim.equal_id (Option.map f) dim access_list)
    | inst -> inst in
    hylo wrap (fn %> unwrap)

  let vectorize_on_dim dim =
    let open InstAlgebra in
    let open Either in
    let select vec scal = function
       (Left e1, Left e2) -> Left (vec e1 e2)
      | (Left e1, Right e2) -> Left (vec e1 (Vbcst e2))
      | (Right e1, Left e2) -> Left (vec (Vbcst e1) e2)
      | (Right e1, Right e2) -> Right (scal e1 e2) in
    let vectorize = function
      | ReadF (tens, acc) when T.does_access tens acc dim ->
        Left (Vread (tens, acc))
      | ReadF (tens, acc) ->
        Right (Read (tens, acc))
      | WriteF (Left expr, tens, acc) when T.does_access tens acc dim ->
        Left (Vwrite (expr, tens, acc))
      | WriteF (Right expr, tens, acc) ->
        Right (Write (expr, tens, acc))
      | WriteF _ ->
        failwith "Trying to write a vector with a scal access"
      | AddF(e1, e2) ->
        select (fun e1 e2 -> Vadd (e1, e2)) (fun e1 e2 -> Add (e1, e2))
          (e1, e2)
      | SubF (e1, e2) ->
        select (fun e1 e2 -> Vsub (e1, e2)) (fun e1 e2 -> Sub (e1, e2))
          (e1, e2)
      | MulF (e1, e2) ->
        select (fun e1 e2 -> Vmul (e1, e2)) (fun e1 e2 -> Mul (e1, e2))
          (e1, e2)
      | AssignF (Left e, Scal, name) ->
        Left (Assign (e, Vec, name))
      | AssignF (Right e, Scal, name) ->
        Right (Assign (e, Scal, name))
      | AssignF _ ->
        failwith "Incompatible vector"
      | AllocatedF (Scal, name) ->
        Right (Allocated (Scal, name))
      | AllocatedF (Vec, name) ->
        Left (Allocated (Vec, name))
      | NopF ->
        Right Nop
      | VaddF _ | VsubF _ | VmulF _ | VbcstF _ | VreadF _ | VwriteF _ ->
        failwith "Already vectorized" in
    hylo wrap vectorize
    %> function Left x | Right x -> x

  let num_allocated_loads = ref 0
  let num_allocated_vloads = ref 0
  let num_allocated_stores = ref 0
  let num_allocated_vstores = ref 0

  type mem_accesses_inf = (T.t * T.accesses) * (inst_type * string)
  (* Finds every access that is independent of all dimensions in dims, factorize them out
   * We have a problem : How to prevent naming interference. We have to keep some
   * track of how many variables have been allocated. No clear idea how to do that now *)
  let fact_loads_on_dims dims prev_accesses inst_list =
    let recurse_inst accesses loads =
      let bin_op op (a1, l1, i1) (a2, l2, i2) =
        List.rev_append a1 a2, List.rev_append l1 l2,
        op i1 i2 in
      let gen_name acc_type loads value tensor expr =
        match List.assoc_eq  [%eq: T.t * (Dim.id * VarExpr.t) list] (tensor, expr) accesses with
        | None ->
          (match List.assoc_eq [%eq: T.t * (Dim.id * VarExpr.t) list] (tensor, expr) prev_accesses with
             None -> (match acc_type with
               | Scal ->
                 let num_load = Ref.post_incr num_allocated_loads in
                 let load_name ="mem_scal_" ^ string_of_int num_load in
                 (((tensor, expr), (Scal, load_name))::accesses),
                 (Assign(value, Scal, load_name)::loads),
                 Allocated(Scal, load_name)
               | Vec ->
                 let num_load = Ref.post_incr num_allocated_vloads in
                 let load_name ="mem_vec_" ^ string_of_int num_load in
                 (((tensor, expr), (Vec, load_name))::accesses),
                 (Assign(value, Vec, load_name)::loads),
                 Allocated(Vec, load_name)
             )
           | Some (_, var) ->
             (((tensor, expr), (acc_type, var))::accesses),
             (Assign(value, acc_type, var)::loads),
             Allocated(acc_type, var)
          )
        | Some (_, var) ->
          accesses, loads, Allocated(acc_type, var) in
      function
      | (ReadF (tensor, accesses) as r)
        when List.for_all (fun dim -> not (T.does_access tensor accesses dim)) dims ->
        gen_name Scal loads (unwrap r) tensor accesses
      | (VreadF (tensor, accesses) as r)
        when List.for_all (fun dim -> not (T.does_access tensor accesses dim)) dims ->
        gen_name Vec loads (unwrap r) tensor accesses
      | inst -> begin match explode inst with
            Binop (v1, v2, f) ->
            bin_op f v1 v2
          | Unop ((accesses, lds, i), f) ->
            accesses, lds, f i
          | No_op i ->
            accesses, [], i
        end in
    let rec factorize accesses_list loads filtered = function
      | [] -> List.rev accesses_list, List.rev loads, List.rev filtered
      | inst::tail ->
        let accesses_list, loads, new_inst = InstAlgebra.hylo wrap (recurse_inst accesses_list loads) inst in
        factorize accesses_list loads (new_inst::filtered) tail in
    factorize [] [] [] inst_list

  let is_tensor_readonly inst_list tensor =
    let does_modify_tens = function
      | Write(_, tens, _) | Vwrite(_, tens, _) when T.equal_id (T.id tens) (T.id tensor) ->
        true
      | _ -> false in
    List.for_all (Bool.not % does_modify_tens) inst_list

  (* Finds every write access that is independent of all dimensions in dim, factorize them out *)
  let fact_stores_on_dims dims prev_accesses inst_list =
    let gen_name acc_type accesses stores filtered value tensor expr =
      match List.assoc_eq  [%eq: T.t * (Dim.id * VarExpr.t) list] (tensor, expr) accesses with
      | None ->
        (match List.assoc_eq [%eq: T.t * (Dim.id * VarExpr.t) list] (tensor, expr) prev_accesses with
           None -> (match acc_type with
             | Scal ->
               let num_stores = Ref.post_incr num_allocated_stores in
               let store_name ="mem_scal_" ^ string_of_int num_stores in
               let allocated = Allocated(Scal, store_name) in
               (((tensor, expr), (Scal, store_name))::accesses),
               (Write(allocated, tensor, expr)::stores),
               (Assign(value, Scal, store_name)::filtered)
             | Vec ->
               let num_stores = Ref.post_incr num_allocated_vstores in
               let store_name = "mem_vec_" ^ string_of_int num_stores in
               let allocated = Allocated(Vec, store_name) in
               (((tensor, expr), (Vec, store_name))::accesses),
               (Vwrite(allocated, tensor, expr)::stores),
               (Assign(value, Vec, store_name)::filtered)
           )
         | Some (_, var) ->
           (((tensor, expr), (acc_type, var))::accesses),
           (let allocated = Allocated(acc_type, var) in
            match acc_type with
              Scal -> Write(allocated, tensor, expr)::stores
            | Vec -> Vwrite(allocated, tensor, expr)::stores),
           (Assign(value, acc_type, var)::filtered)
        )
      | Some (_, var) ->
        accesses, stores, (Assign (value, acc_type, var)::filtered)
    in
    let rec factorize accesses_list stores filtered = function
      | [] -> List.rev accesses_list, List.rev stores, List.rev filtered
      (* There is no need to call recursively factor_stores on
       * instructions because no write instructions must appear in this
       * place *)
      | Write (value, tensor, accesses) ::tail
        when List.for_all (fun dim -> not (T.does_access tensor accesses dim)) dims ->
        let accesses, stores, filtered =
          gen_name Scal accesses_list stores filtered value tensor accesses in
        factorize accesses stores filtered tail
      | Vwrite (value, tensor, accesses)::tail
        when List.for_all (fun dim -> not (T.does_access tensor accesses dim)) dims ->
        let accesses, stores, filtered =
          gen_name Vec accesses_list stores filtered value tensor accesses in
        factorize accesses stores filtered tail
      | inst::tail ->
        factorize accesses_list stores (inst::filtered) tail
    in
    factorize [] [] [] inst_list

  let swap_tensor src_tensor dst_tensor inst_list =
    let swap = function
      | ReadF (tensor, accesses) when T.equal tensor src_tensor ->
        Read (dst_tensor,  accesses)
      | VreadF (tensor, accesses) when T.equal tensor src_tensor ->
        Vread (dst_tensor,   accesses)
      | WriteF (value, tensor, accesses) when T.equal tensor src_tensor ->
        Write (value, dst_tensor,  accesses)
      | VwriteF (value, tensor, accesses) when T.equal tensor src_tensor ->
        Vwrite (value, dst_tensor, accesses)
      | inst -> unwrap inst in
    List.map (InstAlgebra.hylo wrap swap) inst_list

  let swap_accesses f inst_list =
    let swap  = function
      | ReadF (tensor, accesses) ->
        Read (tensor, f tensor accesses)
      | VreadF (tensor, accesses) ->
        Vread (tensor, f tensor accesses)
      | WriteF (value, tensor, accesses) ->
        Write (value, tensor, f tensor accesses)
      | VwriteF (value, tensor, accesses) ->
        Vwrite (value, tensor, f tensor accesses)
      | inst -> unwrap inst in
    List.map (InstAlgebra.hylo wrap swap) inst_list
    
  type fresh = Fresh of string | Old of string

  let gen_code tab value_list =
    let gen_name inst_type state value =
      match value with
      | Allocated (_, name) -> Old name
      | _ ->
        let id, name_list = !state in
        match List.assoc_eq equal value name_list with
        | Some ident ->  Old ident
        | None -> let new_ident = Format.sprintf "%s_%d" (string_of_inst_type inst_type) id in
          state := id + 1, (value, new_ident)::name_list; Fresh new_ident
    and scal_state =ref (0, [])
    and vec_state =ref (0, []) in
    let gen_vec_name = gen_name Vec scal_state
    and gen_scal_name = gen_name Scal vec_state in

    let rec get_code_var gen_name inst_type op = match (gen_name op) with
      | Fresh var ->
        begin match inst_type with
          |Scal ->
            let var_list, vec_list, code = gen_inst op in
            var::var_list, vec_list, var, code
          | Vec ->
            let var_list, vec_list, code = gen_inst op in
            var_list, var::vec_list, var, code
        end
      | Old var ->
        [], [], var, ""

    and format_binop inst_type gen_name format_inst_string tab res op1 op2 =
      let f_a var_a =
        let var_list1, vec_list1, var1, code1 = get_code_var gen_name inst_type op1 in
        let var_list2, vec_list2, var2, code2 = get_code_var gen_name inst_type op2 in
        let var_list = List.concat [var_list1 ; var_list2]
        and vec_list = List.concat [vec_list1 ; vec_list2]
        and code = String.concat "\n" [code1; code2] in
        let op = Printf.sprintf format_inst_string var1 var2 in 
        var_list, vec_list,
        Format.sprintf "%s\n%s%s = %s;" code tab var_a op in
      match gen_name res with
      | Old var_a ->  f_a var_a
      | Fresh var_a ->  let var_list, vec_list, code = f_a var_a in
        begin
          match inst_type with
          |Scal -> var_a::var_list, vec_list, code
          | Vec -> var_list, var_a::vec_list, code
        end

    and format_3op inst_type gen_name format_inst_string tab res op1 op2 op3 =
      let f_a var_a =
        let var_list1, vec_list1, var1, code1 = get_code_var gen_name inst_type op1 in
        let var_list2, vec_list2, var2, code2 = get_code_var gen_name inst_type op2 in
        let var_list3, vec_list3, var3, code3 = get_code_var gen_name inst_type op3 in
        let var_list = List.concat [var_list1 ; var_list2; var_list3]
        and vec_list = List.concat [vec_list1 ; vec_list2; vec_list3]
        and code = String.concat "\n" [code1; code2; code3] in
        let op = Printf.sprintf format_inst_string var1 var2 var3 in 
        var_list, vec_list,
        Format.sprintf "%s\n%s%s = %s;" code tab var_a op in

      match gen_name res with
      | Old var_a ->  f_a var_a
      | Fresh var_a ->
        let var_list, vec_list, code = f_a var_a in
        begin
          match inst_type with
          |Scal -> var_a::var_list, vec_list, code
          | Vec -> var_list, var_a::vec_list, code
        end

    and gen_inst = function
      | Nop  -> [], [], Format.sprintf "%snop;" tab
      | Allocated(Scal, name) -> [name], [], Format.sprintf "%s%s;" tab name
      | Allocated(Vec, name) -> [], [name], Format.sprintf "%s%s;" tab name
      | Assign(Read(tensor, accesses), Scal, name) ->
        [name], [], Format.sprintf "%s%s = %s;" tab name (T.gen_access tensor accesses)
      | Assign(value, Scal, name) ->
        (match gen_scal_name value with
         | Fresh var ->
           let scal_list, vec_list, code = gen_inst value in
           var::name::scal_list, vec_list, Format.sprintf "%s\n%s%s = %s;" code tab name var
         | Old var ->
           [name], [], Format.sprintf "%s%s = %s;" tab name var
        )
      | Read (tensor, accesses) as r ->
        (match gen_scal_name r with
         | Fresh name ->
           let read_str = Format.sprintf "%s = %s;" name (T.gen_access tensor accesses) in
           [name], [], Format.sprintf "%s%s" tab read_str
         | Old name ->
           let read_str = Format.sprintf "%s = %s;" name (T.gen_access tensor accesses) in
           [], [], Format.sprintf "%s%s" tab read_str)
      | Write (value, tensor, accesses)  ->
        (match gen_scal_name value with
         | Fresh var ->
           let scal_list, vec_list, code = gen_inst value in
           let wr_str = Format.sprintf "%s = %s;" (T.gen_access tensor accesses) var in
           var::scal_list, vec_list, Format.sprintf "%s\n%s%s" code tab wr_str
         | Old var ->
           let wr_str = Format.sprintf "%s = %s;" (T.gen_access tensor accesses) var in
           [], [], Format.sprintf "%s%s" tab wr_str
        )
      | Add (v1, v2) as a ->
        format_binop Scal gen_scal_name "%s + %s;" tab a v1 v2
      | Sub (v1, v2) as a ->
        format_binop Scal gen_scal_name "%s - %s;" tab a v1 v2
      | Mul (v1, v2) as a ->
        format_binop Scal gen_scal_name "%s * %s;" tab a v1 v2
      | Vbcst value as vb ->
        (match gen_vec_name vb with
         | Old var_a ->
           (match  gen_scal_name value with
            | Fresh var1 ->
              let scal_list, vec_list, code = gen_inst value in
              let op = Format.sprintf V.gen_broadcast var1 in
              var1::scal_list, vec_list, Format.sprintf "%s\n%s%s = %s;" code tab var_a op
            | Old var1 ->
              let op = Format.sprintf V.gen_broadcast var1 in
              [], [], Format.sprintf "%s%s = %s;" tab var_a op
           )
         | Fresh var_a ->
           (match  gen_scal_name value with
            | Fresh var1 ->
              let var_list, vec_list, code = gen_inst value in
              let op = Format.sprintf V.gen_broadcast var1 in
              var1::var_list, var_a::vec_list,
              Format.sprintf "%s\n%s%s = %s;" code tab var_a op
            | Old var1 ->
              let op = Format.sprintf V.gen_broadcast var1 in
              [], [var_a], Format.sprintf "%s%s = %s;" tab var_a op
           ))
      | Vread (tensor, accesses) as r ->
        (match gen_vec_name r with
         | Fresh name ->
           let op = Format.sprintf V.gen_load (T.gen_access tensor accesses) in
           [], [name], Format.sprintf "%s%s = %s;" tab name op
         | Old name ->
           let op = Format.sprintf V.gen_load (T.gen_access tensor accesses) in
           [], [], Format.sprintf "%s%s = %s;" tab name op)
      | Assign(Vread(tensor, accesses), Vec, name) ->
        let op = Format.sprintf V.gen_load (T.gen_access tensor accesses) in
        [], [name], Format.sprintf "%s%s = %s;" tab name op
      | Assign(value, Vec, name) ->
        (match gen_vec_name value with
         | Fresh var ->
           let scal_list, vec_list, code = gen_inst value in
           scal_list, var::name::vec_list, Format.sprintf "%s\n%s%s = %s;" code tab name var
         | Old var ->
           [], [name], Format.sprintf "%s%s = %s;" tab name var
        )
      | Vwrite (value, tensor, accesses)  ->
        (match gen_vec_name value with
         | Fresh var ->
           let scal_list, vec_list, code = gen_inst value in
           let op = Format.sprintf V.gen_store (T.gen_access tensor accesses) var in
           scal_list, var::vec_list, Format.sprintf "%s\n%s%s;" code tab op
         | Old var ->
           let op = Format.sprintf V.gen_store (T.gen_access tensor accesses) var in
           [], [], Format.sprintf "%s%s;"tab op
        )
      | Vadd (Vmul(v1, v2), v3) | Vadd (v3, Vmul(v1, v2)) as a ->
        format_3op Vec gen_vec_name V.gen_fma tab a v1 v2 v3 
      | Vadd (v1, v2) as a ->
        format_binop Vec gen_vec_name V.gen_add tab a v1 v2
      | Vsub (v1, v2) as s ->
        format_binop Vec gen_vec_name V.gen_sub tab s v1 v2
      | Vmul (v1, v2) as m ->
        format_binop Vec gen_vec_name V.gen_mul tab m v1 v2
    in
    let scal_insts, vec_insts, code = List.fold_left
        (fun (scal, vec, code) inst ->
           let scal', vec', code' = gen_inst inst in List.append scal' scal, List.append vec' vec, code'::code)
        ([], [], [])  value_list in
    List.rev scal_insts, List.rev vec_insts, String.concat "\n" (List.rev code)
end
