open Ids
open Exprs
open Utils
module VE = VarExpr
module CE = ConstExpr
open Instruction


module L(Inst: Inst_t) = struct
  type loop =
    | Once of loop list
    | Statement of Inst.t
    | VLoop of Inst.t (* This loop is guaranteed to contain only vector instructions *)
    | Unroll of {comment: string option; dim_index_id: Index.t; start: VarExpr.t ;
                 size: int; increment: ConstExpr.t; body: loop list}
    | Loop of {comment: string option; dim_index_id: Index.t; aux: (Index.t * VE.t) list; start: VarExpr.t ;
               halt: VarExpr.t; increment: ConstExpr.t; body: loop list} [@@deriving show]


  (* Returns Some list of instructions if all subloops in loop are instructions
   * Returns None otherwise *)
  let to_instructions_list loop_list =
    let rec to_inst_lst acc = function
      | [] -> Some (List.rev acc)
      | Statement inst::tail -> to_inst_lst (inst::acc) tail
      | VLoop inst::tail -> to_inst_lst (inst::acc) tail
      | _ -> None
    in
    to_inst_lst [] loop_list

  let unroll_insts_dim dim start size increment insts =
    let rewrite_insts cnst inst =
      Inst.map_expr (Dim.id dim)
        (VarExpr.alpha_replace_dim dim (VarExpr.(Infix.(start + const cnst * increment )))) inst in
    List.init size Fun.id
    |> List.map (fun cnst -> List.map (rewrite_insts cnst) insts)
    |> List.flatten

  module Rev_loop_graph = struct
    type block = Stmts of Inst.t list [@@deriving show]

    type rev_loop = 
      | Unr of {index: Index.t ; start: VarExpr.t; unroll_size: int; incr: ConstExpr.t ;
                comment: string option; }
      | Seq of {index: Index.t ; aux: (Index.t * VE.t) list; start: VarExpr.t ; halt: VarExpr.t ;
                incr: ConstExpr.t ; comment: string option; } [@@deriving show]

    let new_unroll ?comment index start unroll_size incr =
      Unr { index; start; unroll_size; incr; comment ; }

    let new_seq ?comment ?aux index start halt incr =
      let aux = match aux with Some l -> l | None -> [] in
      Seq { index; start; aux; halt; incr; comment ; }

    let set_comment comment = function
      | Unr args -> Unr {args with comment = Some comment}
      | Seq args -> Seq {args with comment = Some comment}

    module LGraph = Graph.G(struct
        type leaf = block [@@deriving show]
        type node = rev_loop [@@deriving show]
      end)

    include LGraph

    let get_block loop_nest key =
      let Stmts stmts = LGraph.find_leaf key loop_nest in
      stmts

    let top = LGraph.Top

    let is_top key =
      key = top

    type key = LGraph.key [@@deriving show]

    type graph = LGraph.graph  [@@deriving show]

    let empty = LGraph.empty

    let embed_before = embed_before

    let embed_after = embed_after

    let insert_block loop_nest block = 
      let block = Stmts block in
      insert_base block loop_nest

    let modify_block loop_nest b_key f =
      let f = function Stmts block -> Stmts (f block) in
      LGraph.modify_base b_key f loop_nest

    let set_aux loop_nest index_aux_list =
      List.fold_left (fun loop_nest (ind, aux) ->
          LGraph.map_node (function
                Seq ({index;_} as sargs) when Index.equal index ind ->
                Seq {sargs with aux}
              | other -> other) loop_nest) loop_nest index_aux_list

    let set_top_comment loop_nest comment =
      LGraph.modify_node top (set_comment comment) loop_nest

    let hook_block loop_nest enclosing block = 
      let block = Stmts block in
      snd (hook_base block enclosing loop_nest)

    let enclose loop_nest enclosing =
      enclose enclosing loop_nest

    let to_loop loop_nest =
      let rec from_tree = function
        | Leaf (Stmts insts) -> List.map (fun i -> Statement i) insts
        | Top_loop l ->
          let loop_list = List.concat (List.map from_tree l) in
          [Once loop_list]
        | Node (Unr {index; start; unroll_size; incr ;
                     comment } , tree_list) ->
          let subtrees = List.concat (List.map from_tree tree_list) in
          [Unroll {dim_index_id = index; start; size=unroll_size;
                   increment=incr; comment; body = subtrees}]
        | Node (Seq { index ; start; halt; incr; comment; aux; } , tree_list) ->
          let subtrees = List.concat (List.map from_tree tree_list) in
          [Loop {dim_index_id = index; aux; start; halt; increment=incr; body=subtrees; comment}] in
      let tree = to_tree loop_nest in
      Option.get @@ List.hd @@ from_tree tree
  end

  let unroll ?st:(start= VE.Zero)  size dim increment body =
    Unroll {comment = None; dim_index_id = Index.from_dim dim; start; size; increment; body}

  let loop  ?st:(start= VE.Zero) ?halt ?incr:(incr= CE.One) dim body =
    let halt = Option.default (VE.const_expr (CE.size (Dim.size_id dim))) halt in
    Loop {dim_index_id = Index.from_dim dim; aux=[]; start; halt; increment = incr; body; comment = None}

  let rec propagate_expr old_id expr = function
    | Statement stmt -> Statement (Inst.map_expr (Index.dim old_id) (VarExpr.alpha_replace old_id expr) stmt)
    | VLoop vec -> VLoop (Inst.map_expr (Index.dim old_id) (VarExpr.alpha_replace old_id expr) vec)
    | Unroll ({ start; body; _} as l_args) ->
      let start = VarExpr.alpha_replace old_id expr start in
      let body = List.map (propagate_expr old_id expr) body in
      Unroll {l_args with start; body}
    | Loop ({ start; halt; increment; body; _} as l_args) ->
      let map_3tuple f x y z = f x, f y, f z in
      let start, increment, halt = map_3tuple (VarExpr.alpha_replace old_id expr) start increment halt in
      let body = List.map (propagate_expr old_id expr) body in
      Loop {l_args with  start; increment; body; halt}
    | Once loop -> Once (List.map (propagate_expr old_id expr) loop)

  let rec gen_loop_code tab size_indexes loop_nest: string list * string list * string =
    let gen_body body = begin match to_instructions_list body with
        Some inst_list -> Inst.gen_code (tab ^ "\t") inst_list
      | None ->
        let scal_list, vec_list, code_list = List.split3 @@ List.map (gen_loop_code (tab ^ "\t") size_indexes) body in
        List.concat scal_list, List.concat vec_list, String.concat "\n" code_list end in
    match loop_nest with
    | Statement inst -> Inst.gen_code tab [inst]
    | VLoop inst ->Inst.gen_code tab [inst]
    | Once loops -> gen_body loops
    | Unroll { comment; dim_index_id; start; size; body; increment; _;} ->
      let comment = Option.map_default (fun c -> tab ^ " //" ^ c ^ "\n") ""  comment in 
      (* List of numbers from 0 to loop_size - 1 *)
      List.init size Fun.id
      |> List.map (fun cnst_index -> List.map
                      (propagate_expr dim_index_id (VarExpr.(Infix.(start + increment * const cnst_index)))) body)
      |> List.flatten
      |> gen_body
      |> fun (sl, vl, code) -> sl, vl, comment ^ code
    | Loop { comment; dim_index_id; aux; start; halt; increment; body;_;} ->
      let scal_list, vec_list, body_str = gen_body body in
      let index_str = (Index.show_id (Index.id dim_index_id))
      and start_str = (VarExpr.show start)
      and halt_str = (VarExpr.show halt)
      and increment_str = (ConstExpr.show increment) in
      let aux_list = List.map (fun (index, start) -> Printf.sprintf "%s = %s"
                                  (Index.show_id (Index.id index)) (VE.show start)) aux in
      let comment = Option.map_default (fun c -> tab ^ " //" ^ c ^ "\n") ""  comment in 
      let aux_start = match aux_list with [] -> ""
                                        | _ -> ", " ^ String.concat ", " aux_list in
      let aux_list = List.map (fun (index, _) -> Printf.sprintf "%s += %s"
                                  (Index.show_id (Index.id index)) (CE.show increment)) aux in
      let aux_incr = match aux_list with [] -> ""
                                       | _ -> ", " ^ String.concat ", " aux_list in
      scal_list, vec_list, Format.sprintf "%s%sfor (%s = %s%s;\n%s\t%s < %s;\n%s\t%s += %s%s){\n%s\n%s}" comment tab
        index_str start_str aux_start tab index_str halt_str tab index_str increment_str aux_incr body_str tab

  let gen_code sizes_indexes loop_nest =
    let scal_list, vec_list, code = gen_loop_code "" sizes_indexes loop_nest in
    let scal_decl = match scal_list with
        [] -> ""
      | _ -> String.concat " ," (List.sort_unique String.compare scal_list)
             |> Printf.sprintf "%s %s;\n" Inst.A.base_type_name in
    let vec_decl = match vec_list with
        [] -> ""
      | _ -> String.concat " ," (List.sort_unique String.compare vec_list)
             |> Printf.sprintf "%s %s;\n" Inst.A.vec_type_name in
    scal_decl ^ vec_decl ^ code
end

module type Loopnest_t = sig
  module Inst: Inst_t
  type loop =
    | Once of loop list
    | Statement of Inst.t
    | VLoop of Inst.t (* This loop is guaranteed to contain only vector instructions *)
    | Unroll of {comment: string option; dim_index_id: Index.t; start: VarExpr.t ;
                 size: int; increment: ConstExpr.t; body: loop list}
    | Loop of {comment: string option; dim_index_id: Index.t; aux: (Index.t * VE.t) list; start: VarExpr.t ;
               halt: VarExpr.t; increment: ConstExpr.t; body: loop list} [@@deriving show]

  module Rev_loop_graph : sig
    type key [@@deriving show]
    type node [@@deriving show]
    type graph [@@deriving show]

    val top: key
    val is_top: key -> bool

    val empty: graph

    val new_unroll: ?comment:string -> Index.t -> VarExpr.t ->  int -> ConstExpr.t -> node
    val new_seq: ?comment:string -> ?aux:(Index.t * VarExpr.t) list -> Index.t -> VarExpr.t
      -> VarExpr.t -> ConstExpr.t
      -> node

    val insert_block: graph -> Inst.t list -> key * graph
    val modify_block: graph -> key -> (Inst.t list -> Inst.t list) -> graph
    val hook_block: graph  -> key -> Inst.t list -> graph
    val get_block: graph -> key -> Inst.t list

    val set_top_comment: graph -> string -> graph
    val set_aux : graph -> (Index.t * (Index.t * ConstExpr.t) list) list -> graph 

    val enclose: graph -> node -> key * graph
    val to_loop: graph -> loop
    val embed_before: graph -> graph -> key -> graph
    val embed_after: graph -> graph -> key -> graph
  end
  val unroll: ?st:VarExpr.t -> int -> Dim.t -> ConstExpr.t  -> loop list -> loop
  val unroll_insts_dim: Dim.t -> VarExpr.t -> int -> ConstExpr.t -> Inst.t list -> Inst.t list
  val loop: ?st:ConstExpr.t -> ?halt:VarExpr.t -> ?incr:ConstExpr.t -> Dim.t  -> loop list -> loop
  val gen_code: (Size.id * int) list -> loop -> string
end
