open Utils
open Ids
open Exprs
open Loop_nest
module CE = ConstExpr
module VE = VarExpr
module T = Tensor

open Instruction

module PI(Inst: Inst_t)(LN: Loopnest_t with module Inst := Inst)  = struct
  module D = Dim_info.DI(Inst)(LN)
  module RL = LN.Rev_loop_graph
  module DimMap = D.DimMap

  type snapshot = (Dim.id * D.Frozen.t) list

  let show_index_list = [%show: Index.t list]

  type pack = Ready of (Index.t list * RL.key * RL.graph * RL.graph option)
            | Ready_loop of (Index.t list  * RL.graph * RL.graph option)
            | Var_accesses of (RL.key * RL.graph * RL.graph)
            | Var_accesses_loop of (RL.graph * RL.graph)
            | Pending_loop of {tensor: T.t ;
                               load_fn: (T.t -> snapshot -> Index.t list * RL.graph);
                               store_fn: (T.t -> snapshot -> Index.t list * RL.graph) option}
            | Pending of {tensor: T.t ; loop_key: RL.key;
                          load_fn: (T.t -> snapshot -> Index.t list * RL.graph);
                          store_fn: (T.t -> snapshot -> Index.t list * RL.graph) option} [@@deriving show]

  let build_dims_snapshot tensor dim_map =
    let dims = T.dims_list tensor in
    List.map
      (fun dim ->
         Dim.id dim,
         DimMap.find (Dim.id dim) dim_map
         |> D.freeze dim)
      dims

  let build_loop small_snapshot copy_kernel_fn comment pack_tensor big_snapshot =
    let indexes, glob_accesses, loc_accesses, (base_key, loop_nest) = List.fold_left2
        (fun (indexes, glob_accesses, loc_accesses, (base_key, loop_nest))
          (small_dim, small_frozen) (big_dim, big_frozen) ->
          assert (Dim.equal_id small_dim big_dim);
          let new_indexes, glob_index, local_index, dim_loop =
            D.Frozen.interval_loop_tile big_frozen small_frozen in
          List.rev_append new_indexes indexes,
          (small_dim, glob_index)::glob_accesses,
          (small_dim, local_index)::loc_accesses,
          match base_key, dim_loop with
          | None, Some dim_loop -> map_fst (fun a -> Some a) (RL.enclose loop_nest dim_loop)
          | Some key, Some dim_loop -> Some key, snd (RL.enclose loop_nest dim_loop)
          | _, None -> base_key, loop_nest
        )
        ([], [], [], (None, RL.empty)) small_snapshot big_snapshot in
    let base_key = Option.default RL.top base_key in
    indexes, RL.hook_block (RL.set_top_comment loop_nest comment)
      base_key (copy_kernel_fn pack_tensor glob_accesses loc_accesses)

  let push_loop pack_info key =
    List.map (function
          Pending_loop {tensor; load_fn; store_fn} ->
          Pending {tensor; loop_key = key; load_fn; store_fn}
        |Ready_loop (indexes, load, store) ->
          Ready (indexes, key, load, store)
        | Var_accesses_loop ( load, store) ->
          Var_accesses (key, load, store)
        | other -> other) pack_info

  let build_pack_pending dim_map tensor local_tensor is_readonly =
    let load_comment = Printf.sprintf "Pack %s into %s" (T.show_tid tensor) (T.show_tid local_tensor) in
    let store_comment = Printf.sprintf "Unpack %s into %s" (T.show_tid local_tensor)
        (T.show_tid tensor) in
    let dims = T.dims_list tensor in
    let dims_payload = List.filter_map
        (fun dim -> DimMap.find (Dim.id dim) dim_map)
        dims in
    let is_vectorized = List.exists D.is_vectorized  dims_payload in
    let small_snapshot = build_dims_snapshot tensor dim_map in
    let access src_tensor dst_tensor src_accesses dst_accesses = Inst.(
        if is_vectorized
        then [Vwrite( Vread(src_tensor, src_accesses), dst_tensor, dst_accesses)]
        else [Write(  Read(src_tensor, src_accesses), dst_tensor, dst_accesses)]
      ) in
    let load_fn glob_tensor glob_accesses loc_accesses =
      access glob_tensor local_tensor glob_accesses loc_accesses in
    let store_kernel glob_tensor glob_accesses loc_accesses =
      access local_tensor glob_tensor loc_accesses glob_accesses in
    let store_fn = if is_readonly then None
      else Some (build_loop small_snapshot store_kernel store_comment) in
    Pending_loop {tensor;
                  load_fn= build_loop small_snapshot load_fn load_comment;
                  store_fn= store_fn;}

  let apply_dims_pack main_tensor new_pack new_snapshot = function
    | Pending_loop {tensor; load_fn; store_fn;_} when T.equal_id (T.id tensor) (T.id main_tensor) ->
      let indexes, load_loop = load_fn new_pack new_snapshot
      and store_loop = Option.map (fun f -> snd (f new_pack new_snapshot)) store_fn in
      Ready_loop (indexes, load_loop, store_loop)
    | Pending {tensor;loop_key; load_fn; store_fn;_} when T.equal_id (T.id tensor) (T.id main_tensor) ->
      let indexes, load_loop = load_fn new_pack new_snapshot
      and store_loop = Option.map (fun f -> snd (f new_pack new_snapshot)) store_fn in
      Ready (indexes, loop_key, load_loop, store_loop)
    | other -> other

  let apply_tens_to_pending main_tensor new_pack new_snapshot =
    List.map (apply_dims_pack main_tensor new_pack new_snapshot)

  let stack_tensor pack_list dim_map tensor packed_tensor is_readonly =
    let new_snapshot = build_dims_snapshot tensor dim_map in
    let pack_list = apply_tens_to_pending tensor packed_tensor new_snapshot pack_list  in
    let pending = build_pack_pending dim_map tensor packed_tensor is_readonly in
    pending::pack_list

  let stack_var_reduction pack_list load store =
    Var_accesses_loop (load, store)::pack_list

  type final = F of (Index.t list * RL.key * RL.graph * RL.graph option) [@@unboxed]

  type t = (pack list, final list) either

  let empty = Either.Left []

  let flush_waiting_dims _dim_map _main_tensor = function
    | p  -> p

  let flush_tensor_dims pack_list dim_map main_tensor =
    List.map (flush_waiting_dims dim_map main_tensor) pack_list

  let finalize dim_map =
    function
    | Ready (inds, loop_key, load, store) ->
      F (inds, loop_key, load, store)
    | Ready_loop (inds, load, store) ->
      F (inds, RL.top, load, store)
    | Var_accesses ( loop_key, load, store) ->
      F ([], loop_key, load, Some store)
    | Var_accesses_loop (load, store) ->
      F ([], RL.top, load, Some store)
    | Pending {tensor; loop_key; load_fn; store_fn;}  ->
      let last_snapshot = build_dims_snapshot tensor dim_map in
      let inds, load = load_fn tensor last_snapshot in
      let store = Option.map (fun f -> snd (f tensor last_snapshot)) store_fn in
      F (inds, loop_key, load, store)
    | Pending_loop {tensor; load_fn; store_fn;}  ->
      let last_snapshot = build_dims_snapshot tensor dim_map in
      let inds, load = load_fn tensor last_snapshot in
      let store = Option.map (fun f -> snd (f tensor last_snapshot)) store_fn in
      F (inds, RL.top, load, store)

  (* Mapping functions on pack list to apply on t = (pack list, final list) Either.t
   * and to raise CalledOnFinal if they are called on a finalized Pack_info.t *)
  exception CalledOnFinal

  let apply_left f pack_info arg = Either.apply_left_exn CalledOnFinal f pack_info arg
  let apply_left2 f pack_info arg1 arg2 = Either.apply_left2_exn CalledOnFinal f pack_info arg1 arg2
  let apply_left3 f pack_info arg1 arg2 arg3 = Either.apply_left3_exn CalledOnFinal f pack_info arg1 arg2 arg3
  let apply_left4 f pack_info arg1 arg2 arg3 arg4 = Either.apply_left4_exn CalledOnFinal f pack_info arg1 arg2 arg3 arg4

  let push_loop: t -> RL.key -> t = apply_left push_loop

  let stack_tensor pack_info dim_map tensor current_tensor is_readonly =
    apply_left4 stack_tensor pack_info dim_map tensor current_tensor is_readonly

  let stack_var_reduction pack_info load store =
    apply_left2 stack_var_reduction pack_info load store

  let flush_tensor_dims pack_info dim_map tensor =
    apply_left2 flush_tensor_dims pack_info dim_map tensor

  let final_indexes = let open Either in function
      | Right (final_list) -> List.concat (List.map (function F (inds, _, _, _) -> inds) final_list)
      | Left _ -> failwith "List is not finalized yet"

  let final_load_and_store final_list =
    let aux lds_strs = function
        F (_, loop_key, ld_loop, st_loop) -> ( loop_key, ld_loop, st_loop)::lds_strs in
    let pack_loops = List.fold_left aux [] final_list in
    List.rev pack_loops

  let finalize last_dim_map: t -> t = let open Either in function
      | Left pack_list ->
        Right (List.map (finalize last_dim_map) pack_list)
      | Right _ -> failwith "Pack list is already finalized"

  (* Get loads and stores loop *)
  let loads_stores =
    let open Either in function
      | Right final_list -> final_load_and_store final_list
      | Left _ -> failwith "Pack list is already finalized"
end
