open Ids
open Exprs
open Utils

module CE = ConstExpr
module VE = VarExpr

module Args = struct
  let prefix = "T"
end
module T_id = Id(Args)
type id = T_id.t [@@deriving show, ord, eq]
type t = {id: id; dims: Dim.t array; sizes:(Dim.id * CE.t) list;
          strides: (Dim.id * CE.t) list} [@@deriving eq, show]
type gen = unit -> t
type accesses = (Dim.id * VarExpr.t) list [@@deriving eq, show]

let stride tensor dim = Option.get @@ List.assoc_eq Dim.equal_id dim tensor.strides

let dim_size tensor dim = Option.get @@ List.assoc_eq Dim.equal_id dim tensor.sizes

let size {sizes;_} = 
  let non_zero_sizes = List.filter (Bool.not % (CE.equal CE.zero)) (List.map snd sizes) in
  if List.is_empty non_zero_sizes then CE.zero
  else  List.fold_left CE.mul CE.one non_zero_sizes

let id {id;_} = id
let show_tid {id;_} = show_id id

let fresh_gen (): ?name:string -> unit -> id =
  let tensor_gen = T_id.gen_id () in
  fun ?name -> match name with
      Some name -> fun () ->T_id.set_prefix name (tensor_gen ())
    | None -> tensor_gen

let gen_clone ({id;_} as tensor) () = let f_id, gen_id = T_id.stack id () in
  {tensor with id = f_id}, fun () -> {tensor with id = gen_id () }


let dims {dims;_} = dims
let dims_list {dims;_} = Array.to_list dims

let strides_from_size dims dim_size_list =
  List.fold_left (fun (strides, current_stride) dim ->
      let dim_size = Option.default CE.zero (List.assoc dim dim_size_list) in
      if CE.equal CE.zero dim_size then (dim, current_stride)::strides, current_stride
      else let stride = CE.(Infix.( current_stride * dim_size)) in
        (dim, current_stride)::strides, stride)
    ([], CE.one) dims |> fst |> List.rev


(* Create a tensor with dims specified as an array of Dim,
 * with strides and sizes that can be optionally specified,
 * if sizes is not specified, we generate them directly from the dimensions.
 * if strides is not specified, we generate them from size list.
 * In that case, from (I, J, K, L) we generate (1, I, I * J, I * J * K) *)
let make ?name ?strides ?sizes dims  =
  (* For now list are stored in reverse order so that last specified size get stride one *)
  let dims_list = List.rev @@ Array.to_list dims in
  let sizes = match sizes with
    | None -> List.map (fun d -> Dim.id d, CE.size @@ Dim.size_id d) dims_list
    | Some sizes ->
      List.map (fun dim ->
          Dim.id dim,
          Option.default (CE.size @@ Dim.size_id dim) @@ List.assoc (Dim.id dim)
            sizes) dims_list in
  let strides = match strides with
    | None -> strides_from_size (List.map Dim.id dims_list) sizes
    | Some strides -> strides in
  {id = fresh_gen () ?name (); dims; sizes; strides }

let strides_from_size_list tens dim_size_list =
  let dims = List.rev @@ List.map Dim.id @@ dims_list tens in
  strides_from_size dims dim_size_list

let _validate_accesses tensor accesses =
  List.for_all (fun dim -> Array.mem dim tensor.dims) (List.map fst accesses)

let acc_does_access accesses dim =
  List.exists (fun (_, expr) -> VE.access_dim_id expr dim) accesses

(* FIXME: accesses_list seems broken, for now we stay conservative *)
let tens_does_access tensor dim =
  Array.exists (fun d -> Dim.equal_id (Dim.id d) dim) tensor.dims
(* && (List.assoc_opt dim accesses
   |> Option.map VE.is_constant
   |> function Some b -> b
             | None -> false
   )
*)
let does_access tensor accesses dim =
  tens_does_access tensor dim
  || acc_does_access accesses dim

let modify_size_stride tensor size_list =
  let dims = dims_list tensor in
  let sizes =
    List.map (fun dim ->
        Dim.id dim,
        Option.default CE.zero @@ List.assoc (Dim.id dim) size_list)
      dims in
  let strides = strides_from_size_list tensor sizes in
  {tensor with strides; sizes}

let map_stride tensor f =
  let strides = List.map (fun (d, cexpr) -> (d, f cexpr)) tensor.strides in
  {tensor with strides}

let modify_stride tensor dim f =
  let strides = List.modify_assoc Dim.equal_id f dim tensor.strides in
  {tensor with strides}

let modify_accesses accesses dim f =
  List.modify_assoc Dim.equal_id f dim accesses

let gen_access tensor accesses =
  let accesses_expr = List.fold_left
      (fun acc (dim_id, acc_expr) ->
         let stride_dim = stride tensor dim_id in
         VE.(Infix.(acc + (const_expr stride_dim) * acc_expr))
      ) VE.zero accesses in
  Format.sprintf "%s[%s]" (show_id tensor.id) (VE.show accesses_expr)

module OP = struct
  type operator = Add | Mul

  type tensor = t
  type t =
      Tens of tensor
    | Red of operator * Dim.t list * t
    | Join of Dim.t list * t * t

  let tensor tens =
    dims_list tens, Tens tens

  let is_included in_dims out_dims = 
    not (List.for_all (fun dim -> List.exists (Dim.equal dim) out_dims) in_dims)

  let join dims (dims1, a1) (dims2, a2) =
    if not ( is_included dims dims1 && is_included dims dims2)
    then None
    else Some
        (let l1 = List.filter (fun dim -> not @@ List.mem dim dims) dims1 
         and l2 = List.filter (fun dim -> not @@ List.mem dim dims) dims2 in
         List.concat [l1; dims; l2], Join (dims, a1, a2) )

  let reduce operator dims (arg_dims, exp) =
    if not ( is_included dims arg_dims)
    then None
    else Some
      ( let l = List.filter (fun dim -> not @@ List.mem dim dims) arg_dims in
        l, Red (operator, dims, exp) )
end
