open Utils
open Ids
open Exprs
module CE = ConstExpr
module VE = VarExpr
module T = Tensor
open Instruction


module Legacy(Inst: Inst_t) = struct
  type combine = {in_: Tensor.t; out: Tensor.t; map: (Dim.t * Dim.t list) list }
  (* Mapping a const expression in outer tensor index dimension to an outer tensor indexes expressions *)
  module LN = Loop_nest.L(Inst) 

  (* Maps a size from outer tensor to a size in inner tensor *)
  let map_back_sizes {map; _} size_id =
    (* Finding inner sizes list that are mapped to this particular size *)
    List.find_map
      (fun (out_id, in_ids) ->
         if Size.equal_id (Dim.size_id out_id) size_id then Some in_ids
         else None) map
    |> Option.get
    |> List.map Dim.size_id
    |> List.map CE.size
    |> List.reduce (fun red elem -> CE.mul red elem)

  type range = Full | Singl of VE.t | From of VE.t | Until of VE.t | Range of VE.t * VE.t

  let map_range f = function
    | Full -> Full
    | Singl var_expr -> Singl (f var_expr)
    | From var_expr -> From (f var_expr)
    | Until var_expr -> Until (f var_expr)
    | Range (from, to_) -> Range (f from, f to_)

  (* Takes a association list of dim_id and (ranges, loop_type), generates a loop
   * nest that ranges over the specified ranges in the specified dim order
   * We can also specify the type of loop that are used *)
  let visit_range tensor (range_list: (Dim.t * (range * 'a)) list) = let open LN in
    let rec gen_loop access_expr dims range_list = match dims with
      | [] -> Statement Inst.(Read (tensor, access_expr))
      | dim::dims ->
        let range, _ = Option.get @@ List.assoc dim range_list in
        let size_id =  Dim.size_id dim in
        let index_var = Index.from_dim dim in
        let stride = T.stride tensor (Dim.id dim) in
        let access_expr = List.cons (Dim.id dim, VE.(Infix.(const_expr stride *-- index_var))) access_expr in
        let body = [gen_loop access_expr dims range_list] in
        match range with
        | Full ->
          Loop { dim_index_id= Index.from_dim dim; aux = [];
                 start = VE.Zero; halt = VE.size size_id; increment = CE.One;
                 body; comment = None }
        | Singl access ->
          (* We are not going to declare any variable corresponding to this dimension, so
           * we need to replace every occurrence of it in ranges with the access expression *)
          let replace_index = map_range (VE.alpha_replace (Index.from_dim dim) access) in
          let range_list =
            (* Range_list has type Dim.id * (range * LN.loop_type), we want to
             * act on range so we use map_snd % map_fst *)
            range_list
            |> List.map (map_snd (map_fst replace_index))  in
          gen_loop access_expr dims range_list
        | From from ->
          Loop {  dim_index_id= index_var; aux = [];
                  start = from; halt = VE.size size_id; increment = CE.One; body; comment = None}
        | Until to_ ->
          Loop {  dim_index_id= index_var; aux = [];
                  start = VE.Zero; halt = to_; increment = CE.One; body; comment = None }
        | Range (from, to_) ->
          Loop {  dim_index_id= index_var; aux = [];
                  start = from; halt = to_; increment = CE.One; body; comment = None}
    in
    gen_loop [] (T.dims_list tensor) range_list

  (* Default all range that have not been specified to full range *)
  let visit_range_default tensor range_list =
    let def_ranges = List.map
        (fun dim_id -> match List.assoc_opt dim_id range_list with
             None -> (dim_id, (Full, 1))
           | Some (range, ltype) -> (dim_id, (range, ltype))
        )
        (T.dims_list tensor) in
    visit_range tensor def_ranges

  let visitor_loop t =
    visit_range_default t []
end

module type MM_args_t = sig
  val i_dim: Dim.t
  val j_dim: Dim.t
  val k_dim: Dim.t
  val a: Tensor.t
  val b: Tensor.t
  val c_name: string
end

module Tensor_tile(A: Arch.Vec_arch_t) = struct
  module Inst = Instruction.I(A)
  module LN = Loop_nest.L(Inst) 
  module RL = LN.Rev_loop_graph

  type loop_type = U of int * Dim.t [@printer fun fmt (unroll, dim) -> fprintf fmt
                            "U (%d, %s)" unroll (Dim.show_id_of_t dim)]
                 | V of Dim.t [@printer fun fmt  dim -> fprintf fmt
                            "V %s" (Dim.show_id_of_t dim)]
                 | S of int * Dim.t [@printer fun fmt (size, dim) -> fprintf fmt
                            "S (%d, %s)" size (Dim.show_id_of_t dim)]
                 | Pack_tens of T.t [@printer fun fmt tens -> fprintf fmt
                                    "Pack %s" (T.show_tid tens)]
                 | Pack_var of Dim.t list [@printer fun fmt dim_list -> fprintf fmt
                                   "Pack_var [%s]" (String.concat ";" @@ List.map Dim.show_id_of_t dim_list)]
                 | A of Dim.t [@printer fun fmt  dim -> fprintf fmt
                            "A %s" (Dim.show_id_of_t dim)]
  [@@deriving show { with_path = false }]

  type tile_scheme = loop_type list

  let tile_to_string = [%show: loop_type list]

  module Payload = struct

    (* Module holding all necessary tensor information - for now just which clone to use *)
    module Te = struct
      type t = {tensor_clone: T.gen;  tensor_list: (T.t * CE.t) list;}

      let get_new_tens_size new_tens size_list =
        let new_tensor = T.modify_size_stride new_tens @@ List.map (map_fst Dim.id) size_list in
        let new_size = T.size new_tensor in
        new_tensor, new_size

      let make tensor size_list =
        let tens, tensor_clone = T.gen_clone tensor () in
        let tens, size = get_new_tens_size tens size_list in
        {tensor_clone; tensor_list = [tens, size]}

      let pack ({tensor_clone;tensor_list;_} as state) size_list () =
        let new_tens = tensor_clone () in
        let new_tensor, new_size = get_new_tens_size new_tens size_list in
        {state with tensor_list = (new_tensor, new_size)::tensor_list}

      (* Fails if clone was not called before *)
      let current_tensor {tensor_list;_} =
        List.hd tensor_list |> Option.get |> fst

      let decl_alloc_free_tens {tensor_list;_} =
        tensor_list
        |> List.map (fun (t, s) ->
            Printf.sprintf "%s * %s = (%s *)aligned_alloc(%d, sizeof(%s) * %s);"
               Inst.A.base_type_name (T.show_id (T.id t))
               Inst.A.base_type_name (Inst.A.vec_size * 4)
               Inst.A.base_type_name (CE.show s),
            Printf.sprintf "free(%s);"(T.show_id (T.id t))
          )
        |>  List.split
        |> fun (allocs, frees) -> (String.concat "\n" allocs),(String.concat "\n" frees)
    end

    (* Building Map modules needed to hold dimension and tensor infos *)
    module TidMap = Make_map(struct type t = T.id let compare = T.compare_id end)

    module D = Dim_info.DI(Inst)(LN)
    module DimMap = D.DimMap
    module PI = Pack_info.PI(Inst)(LN)

    type t = {dmap: D.t DimMap.t; loop_nest: RL.graph; tmap: Te.t TidMap.t; pack_info: PI.t}

    let empty = {dmap = DimMap.empty; tmap = TidMap.empty; loop_nest = RL.empty; pack_info = PI.empty}

    let apply_dmap f payload = f payload.dmap
                               |> fun dmap -> {payload with dmap}

    let find_dim dim payload = DimMap.find (Dim.id dim) payload.dmap

    let vectorize payload dim =
      payload
      |> apply_dmap
        (DimMap.modify_opt (Dim.id dim)
           (function
               Some _ -> failwith "Dim has already been defined"
             | None -> Some (D.vectorize @@ D.default dim))
        )

    let modify_map_dim dim f payload =
      let dmap, res = DimMap.modify_map dim f payload.dmap in
      {payload with dmap}, res

    (* Is p true for all (dim, dp) in payload *)
    let for_all_dims p payload = DimMap.for_all p payload.dmap

    (* LOOP NEST related info *)
    let push_loop payload loop_key =
      let pack_info = PI.push_loop payload.pack_info loop_key in
      {payload with pack_info}

    let add_inst ({loop_nest;_} as payload) kernel =
      let key, loop_nest = RL.insert_block loop_nest kernel in
      key, {payload with loop_nest}

    let get_insts {loop_nest;_}  key =
      RL.get_block loop_nest key

    let loop_nest {loop_nest;_} =
      loop_nest

    let modify_block f key ({loop_nest;_} as payload) =
      let loop_nest = RL.modify_block loop_nest key f in
      {payload with loop_nest}

    let set_loop_nest payload loop_nest =
      {payload with loop_nest}

    let enclose ({loop_nest;_} as payload) new_loop =
      let new_enclosing, loop_nest = RL.enclose loop_nest new_loop in
      let payload = {payload with loop_nest} in
      let payload = push_loop payload new_enclosing in
      new_enclosing, payload

    (* TILING *)
    let tile_on_dim payload dim tile_size =
      let payload, (cur_index, next_index, incr) =
        modify_map_dim (Dim.id dim)
          (function
            | Some dpayload ->
              let cur_index = D.current_index dpayload in
              let next_index =  D.next_index dpayload in
              let incr = D.incr dpayload in
              Some (D.(bump_index
                       % map_incr (fun incr -> CE.(Infix.(const tile_size * incr))))
                      dpayload
                   ),
              (cur_index, next_index,  incr)
            | None ->
              let dpayload = D.default dim
                             |> D.map_incr (Fun.const @@ CE.const tile_size) in
              let cur_index = D.current_index dpayload in
              let next_index =  D.next_index dpayload in
              Some D.(bump_index dpayload),
              (cur_index, next_index,  CE.one)
          ) payload in
      let start = VE.index next_index
      and halt = VE.(Infix.(min (next_index --+ incr * (const tile_size)) (size (Dim.size_id dim)))) in
      let comment = Printf.sprintf "Tiling dim %s by %d" (Dim.show_id_of_t dim) tile_size in
      let new_loop = RL.new_seq ~comment cur_index start halt incr in
      let loop_key, loop_nest = RL.enclose payload.loop_nest new_loop in
      let payload = {payload with loop_nest} in
      let payload = push_loop payload loop_key in
      payload, loop_key

    (* Same semantic as List.fold_left f init (DimMap.bindings payload.dmap) *)
    let fold_dims f init payload =
      DimMap.fold f payload.dmap init

    (* Get a list of every indexes used in loops generated so far *)
    let all_indexes payload =
      let dim_indexes = fold_dims (fun _ dp ind_list ->
          let aux = D.indexes_list dp
                    |> List.map (fun (ind, _, l) -> ind, l) in
          List.fold_left (fun l (ind, ind_l) -> ind::(List.rev_append (List.map fst ind_l) l))
            ind_list aux) [] payload in
      let pack_indexes = PI.final_indexes payload.pack_info in
      pack_indexes @ dim_indexes

    let fold_tens f init payload =
      TidMap.fold f payload.tmap init

    let pack_tensor ({dmap; pack_info;_} as payload) tensor size_list is_readonly =
      let tmap, (tpayload, is_packed) = TidMap.modify_map (T.id tensor) (function
            Some tpld -> let clone = Te.pack tpld size_list () in
            Some clone, (Te.current_tensor clone, true)
          | None -> let def_tens_pld = Te.make tensor size_list in
            Some def_tens_pld, (Te.current_tensor def_tens_pld, false)
        )
          payload.tmap in
      let payload, current_tensor = {payload with tmap}, tpayload in
      let pack_info = PI.stack_tensor pack_info dmap tensor current_tensor is_readonly in
      {payload with pack_info}, current_tensor, is_packed

    (* TODO find a better name *)
    let add_var_accesses ({pack_info;_} as payload) loads stores=
      let pack_info = PI.stack_var_reduction pack_info loads stores in
      {payload with pack_info}
    let finalize ({dmap; pack_info; _} as payload) =
      let pack_info = PI.finalize dmap pack_info  in
      {payload with pack_info}

    (* Get loads and stores loop. Assumes finalize has been called *)
    let loads_stores payload =
      PI.loads_stores payload.pack_info

  end

  (* Get size of tile on this particular dimension *)
  let dim_tile_size tile_scheme dim =
    let aux current_size = function
      | V d when Dim.equal d dim -> Inst.A.vec_size
      | U (size, d) | S (size, d) when Dim.equal d dim -> current_size * size
      | _ -> current_size in
    List.fold_left aux 1 tile_scheme

  let gen_loop inst loop_tiles =
    let module P = Payload in
    (* We are going to have an helper_list : (dim, increment, index) that
     * specifies the index to use, the increment, etc.*)
    let gen_loop_fold kernel_key (payload, current_loop_key) =
      function
      | V dim ->
        assert (RL.is_top current_loop_key);
        let vect kernel = List.map (Inst.vectorize_on_dim (Dim.id dim)) kernel in
        let payload = P.vectorize payload dim in
        let payload = P.modify_block vect kernel_key payload in
        payload,  current_loop_key
      | U (unroll_size, dim) ->
        (* Here we must not clone the dimension but keep
         * the same so the inner loops can have the right index *)
        let payload, (cur_index, increment) =
          P.modify_map_dim (Dim.id dim) (function
              | Some elem ->
                Some (P.D.map_incr
                        (fun incr -> CE.(Infix.((const unroll_size * incr)))) elem),
                (P.D.current_index elem, P.D.incr elem)
              | None ->
                Some (P.D.map_incr (Fun.const (CE.const unroll_size)) (P.D.default dim)),
                (Index.from_dim dim, CE.one)
            ) payload in
        begin
          if RL.is_top current_loop_key
          then
            let unroll kernel = LN.unroll_insts_dim
                dim (VE.index cur_index) unroll_size increment kernel in
            let payload = P.modify_block unroll kernel_key payload in
            payload, current_loop_key
          else
            let start = VE.index cur_index in
            let comment = (Printf.sprintf "Unroll dim %s by %d" (Dim.show_id_of_t dim) unroll_size) in
            let new_unroll = RL.new_unroll ~comment cur_index start unroll_size increment in
            let new_enclosing, loop_nest = RL.enclose payload.loop_nest new_unroll in
            let payload = P.set_loop_nest payload loop_nest in
            payload,  new_enclosing
        end
      | S (tile_size, dim) ->
        let payload, new_loop =
          P.tile_on_dim payload dim tile_size in
        (*let new_kernel = RL.push_loop kernel new_loop in *)
        payload,  new_loop
      | Pack_var dim_list ->
        let kernel = P.get_insts payload kernel_key in
        let accesses, loads, kernel =
          Inst.fact_loads_on_dims (List.map Dim.id dim_list) [] kernel in
        let _, stores, kernel =
          Inst.fact_stores_on_dims (List.map Dim.id dim_list) accesses kernel in
        let _, loads = RL.insert_block RL.empty loads
        and _, stores = RL.insert_block RL.empty stores in
        let payload = P.modify_block (Fun.const kernel) kernel_key payload in
        let payload = P.add_var_accesses payload loads stores in
        payload, current_loop_key
      | Pack_tens tensor ->
        let dims = T.dims_list tensor in
        (* Retrieve current status of relevant dimensions - dimensions that appear in the tensor *)
        let dims_payload_opt = List.map
            (fun dim -> dim, P.(find_dim dim payload))
            dims in
        (* Filtered dims payloads - that is, only dimensions that concretely
         * appears in a loop nest around the kernel appear there *)
        let dims_payload = List.filter_map
            (fun (d, opt) -> Option.map (fun pl -> d, pl) opt)
            dims_payload_opt in
        (* List of strides for new tensor dimensions *)
        let size_list =
          List.fold_left (fun size_list (dim, dpayload) ->
              let incr = Option.map_default P.D.incr CE.zero dpayload in
              (dim, incr)::size_list
            )
            [] dims_payload_opt in
        let current_pack_indexes_opt = List.map (fun (d, dp) ->
            d, (P.D.current_pack dp))
            dims_payload in
        let accesses      = List.map (function
              d, Some index -> Dim.id d, fun expr ->
                VE.alpha_replace (Index.from_dim d) (VE.index index) expr
            | d, None -> Dim.id d, fun expr ->
                VE.alpha_replace (Index.from_dim d) VE.zero expr
          )
            current_pack_indexes_opt in
        (* TODO: find a cleaner way to do that ?*)
        let is_readonly = Inst.is_tensor_readonly [inst] tensor  in
        let payload, new_tensor, is_packed = P.pack_tensor payload tensor size_list is_readonly in
        let transform insts =
          Inst.swap_tensor tensor new_tensor insts in
        let map_access dim access_expr new_accesses_map =
          let access = Option.default (Fun.const VE.zero) (List.assoc dim new_accesses_map) in
          dim, access access_expr in
        let swap_new_tensor_access insts =
          Inst.swap_accesses (fun other_tensor old_accesses ->
              if T.equal_id (T.id other_tensor) (T.id new_tensor) then
                (
                  List.map (fun (d, old_expr) -> map_access d old_expr accesses) old_accesses
                )
              else (old_accesses))
            insts in
        let pack kernel =
          let kernel = transform kernel in
          if is_packed then kernel
          else  swap_new_tensor_access kernel in
        let payload = P.modify_block pack kernel_key payload in
        payload, current_loop_key
      | A dim ->
        let payload, (cur_index, incr) = P.modify_map_dim (Dim.id dim) (function
            | Some dp ->
              Some (P.D.close dp),
              (P.D.current_index dp, P.D.incr dp)
            | None ->
              Some (P.D.close @@ P.D.default dim),
              (Index.from_dim dim, CE.one)
          )
            payload in
        let start = VE.zero
        and halt = VE.const_expr (CE.size (Dim.size_id dim)) in
        let new_loop = RL.new_seq cur_index start halt incr in
        let loop_key, payload = P.enclose payload new_loop in
        payload, loop_key  in
    (*End aux function *)

    let kernel_key, payload = P.add_inst P.empty [inst] in
    let payload, _ = List.fold_left (gen_loop_fold kernel_key)
        (payload, RL.top) loop_tiles in
    if not (P.for_all_dims (fun _ elem -> P.D.is_closed elem) payload)
    then failwith "Some dimension is not closed"
    else
      let payload = P.finalize payload in
      let tensor_decl, tensor_free =
        P.fold_tens (fun _ tp (list_decl, list_free) ->
            let decl, free = P.Te.decl_alloc_free_tens tp in
            decl::list_decl, free::list_free) ([], []) payload in
      let tensor_decl = Printf.sprintf "%s\n" (String.concat "\n" tensor_decl) in
      let tensor_free = Printf.sprintf "\n%s\n" (String.concat "\n" tensor_free) in
      let guards = P.fold_dims (fun _ dp buf ->
          let last_incr = Option.get @@ P.D.last_tile_size dp in
          let dim_size = P.D.dim dp |> Dim.size_id in
          let div_guard = if CE.equal CE.zero last_incr then "" else
              Printf.sprintf" && (%s %% %s == 0)" (Size.show_id dim_size)
                (CE.show last_incr) in
          Printf.sprintf "%sassert((%s <= %s)%s);\n" buf
            (CE.show last_incr) (Size.show_id dim_size) div_guard)
          "" payload in

      let list_indexes = P.all_indexes payload |> List.sort_unique Index.compare in
      let indexes_decl = String.concat ", " @@ List.map (Index.show_id % Index.id) list_indexes
                         |> Printf.sprintf "int %s;\n" in
      let loop_nest =
        P.fold_dims (fun _ dp loop_nest ->
            let aux = P.D.indexes_list dp
                      |> List.map (fun (ind, _, l) -> ind, l) in
            RL.set_aux loop_nest aux
          ) (P.loop_nest payload) payload in
      let lds_strs = P.loads_stores payload in
      let loop_nest = List.fold_left (fun loop_nest (l_key, ld, str) ->
          let loop_nest = RL.embed_before ld loop_nest l_key in
          Option.map_default (fun str -> RL.embed_after str loop_nest l_key)
            loop_nest str
        )
          loop_nest lds_strs in
      (Printf.sprintf "/*\n%s\n*/\n" ([%show: loop_type list]  loop_tiles)),
      indexes_decl, tensor_decl, tensor_free, guards, RL.to_loop loop_nest


  module MM_build(MM_args: MM_args_t) = struct
    include MM_args

    let c = T.make ~name:c_name [|i_dim; j_dim|]

    module Tiles = struct
      let blis_tile =
        [ V j_dim;  U(6, i_dim); U (2, j_dim); S (256, k_dim); Pack_var [k_dim];
          S (32, i_dim); Pack_tens b; S (256, j_dim);  Pack_tens a;
          A k_dim; A j_dim; A i_dim]

      let motoko_tile =
        [V j_dim; U(2, j_dim); U (6, i_dim); S (192, k_dim); Pack_var [k_dim];
         Pack_tens b; S (12, j_dim); Pack_tens a;  A i_dim; A k_dim; A j_dim]
    end

    let loop_body = let open Inst in
        let from_dim dim = Dim.id dim, VE.index @@ Index.from_dim dim in
        let read_out = Read (c, [from_dim i_dim; from_dim j_dim])
        and read_1 = Read (a, [from_dim i_dim; from_dim k_dim])
        and read_2 = Read (b, [from_dim k_dim; from_dim j_dim]) in
        let contract = Add (read_out, Mul (read_1, read_2)) in
        Write (contract, c, [from_dim i_dim; from_dim j_dim])

    (* Create a 6*8 gemm kernel on a and b *)
    let gen_code tile_scheme =
      let tile_scheme, indexes_decl, tensor_decl, tensor_free, guards, loop =
        gen_loop loop_body tile_scheme in
      let loop_code = LN.gen_code [] loop in
      let buffer = Buffer.create 2000 in
      let () = List.iter (Buffer.add_string buffer)
          [tile_scheme; indexes_decl; tensor_decl; guards; loop_code; tensor_free] in
      Buffer.contents buffer
  end
end

